import json

from application.interactor.apply.apply_coupon import ApplyCouponInteractor
from application.interactor.apply.apply_coupon.dto.request import ApplyCouponRequest
from application.interactor.apply.dto import ApplyContext
from .._BaseApplyProcessor import BaseApplyProcessor


class ApplyCouponProcessor(BaseApplyProcessor):

    def __init__(
        self,        # type: ApplyCouponProcessor
        interactor,  # type: ApplyCouponInteractor
    ):
        # type: (...) -> None

        super(ApplyCouponProcessor, self).__init__()
        self.interactor = interactor

    def parse_data(
        self,        # type: ApplyCouponProcessor
        data,        # type: bytes
        event_type,  # type: str
    ):
        # type: (...) -> (int, int, str, int)

        parsed_data = json.loads(data)
        pos_id = int(parsed_data.get("posId"))
        originator_pos_id = int(parsed_data.get("originatorPosId") or pos_id)
        coupon_id = str(parsed_data["couponId"])
        user_auth_level = parsed_data.get("user_auth_level")
        if user_auth_level is None:
            user_auth_level = -1

        return ApplyCouponRequest(
            coupon_id=coupon_id,
            context=ApplyContext(
                pos_id=pos_id,
                originator_pos_id=originator_pos_id,
                user_auth_level=user_auth_level,
                applier=None,
            )
        )

    def call_business(
            self,     # type: ApplyCouponProcessor
            request,  # type: ApplyCouponRequest
    ):
        # type: (...) -> None

        self.interactor.apply_coupon(apply_coupon_request=request)



from posot import OrderTaker

from model.error_type import OperationErrorsType
from model.exception import OperationErrors
from model.services import ElectronicPaymentServiceModel
from model.tokens import BusTokens
from typing import (
    Optional,
    List,
)
from xml.etree import cElementTree as eTree

from sysactions import get_used_service
from logging import Logger

from mbcontextmessagehandler import MbContextMessageBus
from messagebus import (
    Message,
)
from msgbus import (
    TK_SYS_ACK,
)


class ElectronicPaymentService(ElectronicPaymentServiceModel):
    def __init__(
        self,
        logger,                     # type: Logger
        message_bus,                # type: MbContextMessageBus
        electronic_payments_id,     # type: Optional[List[int]]
    ):
        # type: (...) -> None

        self.logger = logger
        self.message_bus = message_bus
        self.electronic_payments_id = electronic_payments_id

    def finish_transaction(
        self,
        model,                  # type: eTree
        pos_ot,                 # type: OrderTaker
        electronic_type_id,     # type: int
        pos_id,                 # type: str
        status,                 # type: str
        order_id="",            # type: Optional[str]
    ):
        # type: (...) -> Optional[Message]

        if electronic_type_id not in self.electronic_payments_id:
            return

        order_id = "" if order_id is None else order_id
        order = pos_ot.orderPicture(
            posid=pos_id,
            orderid=order_id,
        )

        if order is None:
            return

        order = eTree.XML(order).find("Order")
        created_at = order.get("createdAt").replace("-", "").replace(":", "")
        fiscal_date = created_at[:8]
        fiscal_time = created_at[9:15]
        order_id = order.get("orderId") if order_id == "" else order_id

        data = ";".join([str(pos_id), order_id, fiscal_date, fiscal_time, status, "", ""])

        payment_service = self.get_electronic_payment_service(model=model)

        for try_num in range(1, 4):
            try:

                msg = "[finish_eft_transactions] Trying to send a finish message. Order #: {}; Try #: {}; STATUS #: {}"
                self.logger.info(msg.format(order_id, try_num, status))

                message = Message(
                    token=BusTokens.TK_ELECTRONIC_PAYMENT_FINISH_PAYMENT,
                    timeout=600000000,
                    data=data,
                )
                response = self.message_bus.send_message(
                    component=payment_service,
                    message=message,
                )

                if response.token == TK_SYS_ACK:
                    msg = "[finish_eft_transactions] Success finished eft transaction. Order #: {}"
                    self.logger.info(msg.format(order_id))
                    return response

                message = "[finish_eft_transactions] ACK not received. Order #: {} - {} - {}"
                self.logger.info(message.format(order_id, response.token, response.data))

            except Exception as _:
                msg = "[finish_eft_transactions] Error sending message. Order #: {} / Try #: {} / Status: {}"
                msg = msg.format(order_id, try_num, status)
                if try_num == 3:
                    self.logger.exception(msg, exc_info=True)
                else:
                    self.logger.info(msg)

        return None

    @staticmethod
    def get_electronic_payment_service(
        model,      # type: eTree
    ):
        # type: (...) -> str

        payment_service = get_used_service(
            model=model,
            service_type="eft",
        )

        if payment_service:
            return payment_service
        raise OperationErrors(OperationErrorsType.InvalidPaymentService)




from _LogoutSummaryInfo import LogoutSummaryInfo
from _LogoutSummaryDto import LogoutSummaryDto
from _LogoutSummaryHeaderDto import LogoutSummaryHeaderDto
from _LogoutSummaryBodyDto import LogoutSummaryBodyDto
from _CashReportBodyDto import CashReportBodyDto
from _CashReportBodyLineDto import CashReportBodyLineDto
from _OrderInfo import OrderInfo

__all__ = ["LogoutSummaryInfo",
           "LogoutSummaryDto",
           "LogoutSummaryHeaderDto",
           "LogoutSummaryBodyDto",
           "CashReportBodyDto",
           "CashReportBodyLineDto",
           "OrderInfo"]



from datetime import datetime, timedelta
from xml.etree import cElementTree as eTree

import iso8601
from application.model import (
    BenefitControllerDto,
    OperationStatus,
    BenefitAppliers,
    OperationDescription,
)
from application.model.configuration import Configurations
from mbcontextmessagehandler import MbContextMessageBus
from mwhelper import BaseRepository
from persistence import Connection
from typing import List, Optional


class DBRepository(BaseRepository):

    def __init__(self, configs, message_bus):
        # type: (Configurations, MbContextMessageBus) -> None

        super(DBRepository, self).__init__(message_bus.mbcontext)

        self.logger = configs.logger
        self.configs = configs
        self.message_bus = message_bus

    def manage_operation_status(
        self,                       # type: DBRepository
        success,                    # type: bool
        benefit_id,                 # type: str
        operation_description,      # type: OperationDescription
        benefit_applier,            # type: str
        retry_quantity,             # type: int
        order_picture=None,         # type: Optional[eTree]
        table_picture=None,         # type: Optional[eTree]
    ):
        # type: (...) -> None

        if success:
            self.delete_pending_operation(benefit_id)
        else:
            operation_status = OperationStatus.NOT_DONE

            retry_quantity += 1
            if retry_quantity > self.configs.max_retry_quantity:
                operation_status = OperationStatus.ERROR
                message = "The benefit has exceeded the max retry quantity. CouponId: {}; RetryQuantity: {}"
                self.logger.info(message.format(benefit_id, retry_quantity))

            self.insert_or_update_pending_operation(
                benefit_id=benefit_id,
                operation_description=operation_description,
                operation_status=operation_status,
                benefit_applier=benefit_applier,
                retry_quantity=retry_quantity,
                order_picture=order_picture,
                table_picture=table_picture,
            )

    def find_pending_operations(
        self,
    ):
        # type: () -> Optional[List[BenefitControllerDto]]
        def inner_func(conn):
            # type: (Connection) -> Optional[List[BenefitControllerDto]]

            query = """
                SELECT * 
                FROM PendingOperations 
                WHERE OperationStatusId = '{}'
            """.format(OperationStatus.NOT_DONE)

            cursor = conn.select(query)
            if not cursor.rows():
                return

            results = []
            for result in cursor:
                benefit = BenefitControllerDto(
                    benefit_id=result.get_entry("BenefitId"),
                    operation_description=int(result.get_entry("OperationDescriptionId")),
                    benefit_applier=result.get_entry("BenefitApplier"),
                    retry_quantity=int(result.get_entry("RetryQuantity")),
                    last_retry_date_utc=iso8601.parse_date(result.get_entry("LastRetryDateUTC")),
                    order_picture=result.get_entry("OrderPicture"),
                    table_picture=result.get_entry("TablePicture"),
                )

                results.append(benefit)

            return results

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def get_benefit_applier_by_db(self, benefit_id):
        # type: (str) -> str

        def inner_func(conn):
            # type: (Connection) -> Optional[BenefitAppliers]

            query = """
                SELECT BenefitApplier 
                FROM PendingOperations 
                WHERE BenefitId = '{0}'
            """.format(benefit_id)

            cursor = conn.select(query)
            if not cursor.rows():
                return

            return cursor.get_row(0).get_entry(0)

        return self.execute_with_connection(inner_func, service_name=self.configs.persistence_name)

    def insert_or_update_pending_operation(
        self,                       # type: str
        benefit_id,                 # type: str
        operation_description,      # type: OperationDescription
        operation_status,           # type: OperationStatus
        benefit_applier,            # type: BenefitAppliers
        retry_quantity,             # type: int
        order_picture=None,         # type: Optional[eTree]
        table_picture=None,         # type: Optional[eTree]
    ):
        # type: (...) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            last_retry_date = datetime.utcnow().isoformat()

            query = """
                INSERT OR REPLACE INTO PendingOperations(
                    BenefitId,
                    OrderPicture,
                    TablePicture,
                    OperationDescriptionId,
                    OperationStatusId,
                    BenefitApplier,
                    RetryQuantity,
                    LastRetryDateUTC
                )
                VALUES ('{0}', {1}, {2}, '{3}', '{4}', '{5}', '{6}', '{7}')
            """
            query = query.format(
                benefit_id,
                "'{}'".format(eTree.tostring(order_picture)) if order_picture else "NULL",
                "'{}'".format(eTree.tostring(table_picture)) if table_picture else "NULL",
                operation_description,
                operation_status,
                benefit_applier,
                retry_quantity,
                last_retry_date,
            )
            conn.query(query)

        self.execute_with_transaction(inner_func, service_name=self.configs.persistence_name)

    def change_benefit_operation_status_to_error(
        self,                       # type: str
        benefit_id,                 # type: str
    ):
        # type: (...) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            query = """
                UPDATE PendingOperations
                SET OperationStatusId = {}
                WHERE BenefitId = {}
            """.format(
                OperationStatus.ERROR,
                benefit_id
            )
            conn.query(query)

        self.execute_with_transaction(inner_func, service_name=self.configs.persistence_name)

    def delete_pending_operation(
        self,            # type: DBRepository
        benefit_id,      # type: str
    ):
        # type: (str) -> None

        def inner_func(conn):
            # type: (Connection) -> None

            query = """
                DELETE FROM PendingOperations 
                WHERE BenefitId = '{0}'
            """.format(benefit_id)
            conn.query(query)

        self.execute_with_transaction(inner_func, service_name=self.configs.persistence_name)



import logging
import os
import tempfile
import shutil
import json
from subprocess import check_call

from dateutil.zoneinfo import tar_open, METADATA_FN, ZONEFILENAME


def rebuild(filename, tag=None, format="gz", zonegroups=[], metadata=None):
    """Rebuild the internal timezone info in dateutil/zoneinfo/zoneinfo*tar*

    filename is the timezone tarball from ftp.iana.org/tz.

    """
    tmpdir = tempfile.mkdtemp()
    zonedir = os.path.join(tmpdir, "zoneinfo")
    moduledir = os.path.dirname(__file__)
    try:
        with tar_open(filename) as tf:
            for name in zonegroups:
                tf.extract(name, tmpdir)
            filepaths = [os.path.join(tmpdir, n) for n in zonegroups]
            try:
                check_call(["zic", "-d", zonedir] + filepaths)
            except OSError as e:
                _print_on_nosuchfile(e)
                raise
        # write metadata file
        with open(os.path.join(zonedir, METADATA_FN), 'w') as f:
            json.dump(metadata, f, indent=4, sort_keys=True)
        target = os.path.join(moduledir, ZONEFILENAME)
        with tar_open(target, "w:%s" % format) as tf:
            for entry in os.listdir(zonedir):
                entrypath = os.path.join(zonedir, entry)
                tf.add(entrypath, entry)
    finally:
        shutil.rmtree(tmpdir)


def _print_on_nosuchfile(e):
    """Print helpful troubleshooting message

    e is an exception raised by subprocess.check_call()

    """
    if e.errno == 2:
        logging.error(
            "Could not find zic. Perhaps you need to install "
            "libc-bin or some other package that provides it, "
            "or it's not in your PATH?")



from .dto.request import ApplyBenefitRequest
from ..service import BenefitApplierService
from ....model.repository.benefit_applier.exception import BenefitApplierNotFound
from ....repository.pos.utils import get_benefit_applier_by_name


class ApplyBenefitInteractor(object):

    def __init__(
        self,                           # type: ApplyBenefitInteractor
        benefit_applier_service,        # type: BenefitApplierService
    ):
        # type: (...) -> None

        self.benefit_applier_service = benefit_applier_service

    def apply_benefit(
        self,                   # type: ApplyBenefitInteractor
        apply_benefit_request,  # type: ApplyBenefitRequest
    ):
        # type: (...) -> None

        self._validate_benefit_applier(apply_benefit_request)

        self.benefit_applier_service.apply_benefit(
            benefit=apply_benefit_request.benefit,
            context=apply_benefit_request.context,
            benefit_applier=apply_benefit_request.context.applier,
        )

    @staticmethod
    def _validate_benefit_applier(apply_benefit_request):
        if apply_benefit_request.context.applier is None:
            raise BenefitApplierNotFound("")
        benefit_applier = get_benefit_applier_by_name(apply_benefit_request.context.applier)
        if benefit_applier is None:
            raise BenefitApplierNotFound(apply_benefit_request.context.applier)



from basereport._AlignTypes import AlignTypes


class ReportColumnDefinition(object):
    def __init__(
            self,
            text='',
            width=38,
            fill_with_char=' ',
            before_text='',
            after_text='',
            align=AlignTypes.LEFT,
            after_fill_text='',
            before_fill_text=''):
        # type: (str, int, unicode, str, str, AlignTypes, str, str) -> None
        self.text = text
        self.width = width
        self.fill_with_char = fill_with_char
        self.before_text = before_text
        self.after_text = after_text
        self.align = align
        self.after_fill_text = after_fill_text
        self.before_fill_text = before_fill_text



from xml.etree import cElementTree as eTree

from application.interactor import BurnBenefitInteractor
from application.model import BenefitAppliers
from application.model.configuration import Configurations
from application.model.repository.benefit_applier import DefaultBenefitApplierRepository
from application.repository.pos import DBRepository
from application.repository.pos.utils import find_order_benefits_id
from typing import Dict


class OrderModifiedPaidInteractor(object):
    def __init__(
        self,                               # type: OrderModifiedPaidInteractor
        burn_benefit_interactor,            # type: BurnBenefitInteractor
        configs,                            # type: Configurations
        db_repository,                      # type: DBRepository
        benefit_appliers_repositories,      # type: Dict[BenefitAppliers: DefaultBenefitApplierRepository]
    ):
        # type: (...) -> None

        self.configs = configs
        self.logger = self.configs.logger
        self.db_repository = db_repository
        self.benefit_appliers_repositories = benefit_appliers_repositories
        self.burn_benefit_interactor = burn_benefit_interactor

    def burn_paid_benefit(
        self,                   # type: OrderModifiedPaidInteractor
        current_pos_id,         # type: int
        originator_pos_id,      # type: int
        order_id,               # type: int
        order_picture,          # type: eTree
    ):
        # type: (...) -> None

        order_benefits_id = find_order_benefits_id(self.logger, order_picture)
        if not order_benefits_id:
            return

        message = "Benefits founded to burn: {}; Current/Originator PosId: {}/{}; OrderId: {}"
        self.logger.info(message.format(order_benefits_id, current_pos_id, originator_pos_id, order_id))

        for benefit_id in order_benefits_id:
            benefit_applier = self.db_repository.get_benefit_applier_by_db(benefit_id)
            if benefit_applier not in [None, 'None', '']:
                self.burn_benefit_interactor.burn_benefit(
                    benefit_id,
                    originator_pos_id,
                    benefit_applier,
                    order_picture
                )



from abc import ABCMeta, abstractmethod


class PosController(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def open_user(self, user_id):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def log_in_user(self, user_id):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def log_out_user(self):
        # type: () -> None
        raise NotImplementedError()

    @abstractmethod
    def close_user(self, user_id):
        # type: (str) -> None
        raise NotImplementedError()

    @abstractmethod
    def set_pod_function(self, pos_function):
        # type: (str) -> None
        raise NotImplementedError()



from commons.report import Generator
from reports_app.loginsummary.dto import LoginSummaryInfo  # noqa
from reports_app.loginsummary.dto import LoginSummaryDto  # noqa
from reports_app.loginsummary.dto import LoginSummaryHeaderDto  # noqa
from reports_app.loginsummary.dto import LoginSummaryBodyDto


class LoginGenerator(Generator):
    def __init__(self, login_summary_info):
        # type: (LoginSummaryInfo) -> ()
        self.login_summary_info = login_summary_info

    def generate_data(self):
        # type: (LoginSummaryInfo) -> LoginSummaryDto
        login_summary_header_dto = LoginSummaryHeaderDto(
            self.login_summary_info.pos_id,
            self.login_summary_info.store_id,
            self.login_summary_info.operator_id,
            self.login_summary_info.business_date
        )
        login_summary_body_dto = LoginSummaryBodyDto(
            self.login_summary_info.initial_value,
            self.login_summary_info.pos_type,
            self.login_summary_info.authorizer
        )

        return LoginSummaryDto(login_summary_header_dto, login_summary_body_dto)



from ._CommonProcessor import _CommonProcessor
from model import ConfirmFinishPaymentData


class ConfirmFinishPaymentProcessor(_CommonProcessor):
    def parse_data(self, data, event_type):
        return ConfirmFinishPaymentData(*data.split(';'))

    def call_business(
        self,
        confirm_finish_payment_data     # type: ConfirmFinishPaymentData
    ):
        # type: (...) -> str

        processor = self.payment_router.get_payment_processor()
        if processor:
            return processor.confirm_finish_payment(confirm_finish_payment_data=confirm_finish_payment_data)



# This file is dual licensed under the terms of the Apache License, Version
# 2.0, and the BSD License. See the LICENSE file in the root of this repository
# for complete details.

from __future__ import absolute_import, division, print_function


from cryptography import utils
from cryptography.exceptions import UnsupportedAlgorithm, _Reasons
from cryptography.hazmat.primitives import hashes


@utils.register_interface(hashes.HashContext)
class _HashContext(object):
    def __init__(self, backend, algorithm, ctx=None):
        self._algorithm = algorithm

        self._backend = backend

        if ctx is None:
            ctx = self._backend._lib.Cryptography_EVP_MD_CTX_new()
            ctx = self._backend._ffi.gc(
                ctx, self._backend._lib.Cryptography_EVP_MD_CTX_free
            )
            evp_md = self._backend._evp_md_from_algorithm(algorithm)
            if evp_md == self._backend._ffi.NULL:
                raise UnsupportedAlgorithm(
                    "{} is not a supported hash on this backend.".format(
                        algorithm.name),
                    _Reasons.UNSUPPORTED_HASH
                )
            res = self._backend._lib.EVP_DigestInit_ex(ctx, evp_md,
                                                       self._backend._ffi.NULL)
            self._backend.openssl_assert(res != 0)

        self._ctx = ctx

    algorithm = utils.read_only_property("_algorithm")

    def copy(self):
        copied_ctx = self._backend._lib.Cryptography_EVP_MD_CTX_new()
        copied_ctx = self._backend._ffi.gc(
            copied_ctx, self._backend._lib.Cryptography_EVP_MD_CTX_free
        )
        res = self._backend._lib.EVP_MD_CTX_copy_ex(copied_ctx, self._ctx)
        self._backend.openssl_assert(res != 0)
        return _HashContext(self._backend, self.algorithm, ctx=copied_ctx)

    def update(self, data):
        data_ptr = self._backend._ffi.from_buffer(data)
        res = self._backend._lib.EVP_DigestUpdate(
            self._ctx, data_ptr, len(data)
        )
        self._backend.openssl_assert(res != 0)

    def finalize(self):
        if isinstance(self.algorithm, hashes.ExtendableOutputFunction):
            # extendable output functions use a different finalize
            return self._finalize_xof()
        else:
            buf = self._backend._ffi.new("unsigned char[]",
                                         self._backend._lib.EVP_MAX_MD_SIZE)
            outlen = self._backend._ffi.new("unsigned int *")
            res = self._backend._lib.EVP_DigestFinal_ex(self._ctx, buf, outlen)
            self._backend.openssl_assert(res != 0)
            self._backend.openssl_assert(
                outlen[0] == self.algorithm.digest_size
            )
            return self._backend._ffi.buffer(buf)[:outlen[0]]

    def _finalize_xof(self):
        buf = self._backend._ffi.new("unsigned char[]",
                                     self.algorithm.digest_size)
        res = self._backend._lib.EVP_DigestFinalXOF(
            self._ctx, buf, self.algorithm.digest_size
        )
        self._backend.openssl_assert(res != 0)
        return self._backend._ffi.buffer(buf)[:self.algorithm.digest_size]



from ._ApiRequest import ApiRequest
from ._HttpMessageProcessor import HttpMessageProcessor
from ._HttpMessageProcessorMessageHandler import HttpMessageProcessorMessageHandler
from ._RouteNotMapped import RouteNotMapped



# -*- coding: utf-8 -*-

import json

from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import ParseError

import sysactions

from messagebus import (
    DataType,
    DefaultToken,
    Message,
    MessageBus,
)
from msgbus import TK_SLCTRL_OMGR_ORDERPICT
from posot import (
    OrderTaker,
    OrderTakerException,
)

from application.model import (
    Configurations,
    NDiscountModelJsonEncoder,
)
from application.model.benefit import BenefitList
from application.model.exceptions import (
    OrderNotFoundException,
    OrderPictureException,
)
from application.model.order import Order
from application.repository import NDiscountRepository


class OrderService(object):
    def __init__(self, configs, repository, message_bus):
        # type: (Configurations, NDiscountRepository, MessageBus) -> None

        self.configs = configs
        self.logger = configs.logger
        self.repository = repository
        self.message_bus = message_bus

    @staticmethod
    def get_burnt_items(applied_benefits):
        # type: (eTree) -> Dict

        burnt_items = {}
        for benefit in applied_benefits:
            current_burnt_items = benefit["burnt_items"]
            for key in current_burnt_items.keys():
                if key in burnt_items:
                    burnt_items[key] += current_burnt_items[key]
                else:
                    burnt_items[key] = current_burnt_items[key]

        return burnt_items

    def get_applied_benefits_from_order_picture(
        self,
        pos_id,                                     # type: int
        order_picture=None,                         # type: Optional[eTree]
    ):
        # type: (...) -> List

        order_benefits = self.get_benefits_from_order_picture(pos_id, order_picture)
        if not order_benefits:
            return []

        return json.loads(order_benefits.get("benefitList", "[]"))

    def get_benefit_pre_total_sequence_from_order_picture(
        self,                   # type: OrderService
        pos_id,                 # type: int
        order_picture=None,     # type: Optional[eTree]
    ):
        # type: (...) -> List

        order_benefits = self.get_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        if not order_benefits:
            return []

        return order_benefits.get("benefitPreTotalSequence", [])

    def get_benefits_from_order_picture(self, pos_id, order_picture=None):
        # type: (int, Optional[eTree]) -> Dict

        try:
            if order_picture is None:
                order_picture = self.get_order_picture(pos_id)

            value = order_picture.find(".//OrderProperty[@key=\"BENEFIT_LIST\"]").get("value")
            return json.loads(value)
        except OrderTakerException:
            pass
        except AttributeError:
            pass
        return {}

    @staticmethod
    def save_benefit_list_custom_property(
        pos_ot,         # type: OrderTaker
        benefit_list,   # type: Union[BenefitList, str]
    ):
        # type: (...) -> Optional[eTree.Element]

        if benefit_list == "clear":
            benefit_list_json = benefit_list
        else:
            benefit_list_json = json.dumps(benefit_list, encoding="utf-8", cls=NDiscountModelJsonEncoder)

        order_xml = pos_ot.setOrderCustomProperty("BENEFIT_LIST", benefit_list_json)

        try:
            return eTree.XML(order_xml)
        except (TypeError, ParseError):
            return None

    def get_order_picture(
        self,
        pos_id,         # type: Any[Union[str, int]]
        order_id=None,  # type: Optional[str]
    ):
        # type: (...) -> Optional[eTree.Element]

        data = "{}\0{}".format(str(pos_id), str(order_id or ''))
        message = Message(TK_SLCTRL_OMGR_ORDERPICT, DataType.param, data, 30000000)
        reply = self.message_bus.send_message("ORDERMGR{}".format(str(pos_id)), message)

        if reply.token == DefaultToken.TK_SYS_ACK:
            try:
                split_reply = reply.data.split("\0")
                if not split_reply[2]:
                    return None
                order_picture = eTree.XML(split_reply[2])
                if order_id:
                    order_picture = order_picture.findall('.//Order')[0]
                return order_picture
            except AttributeError:
                return None

        raise OrderPictureException("Error getting OrderPicture: {} - {}".format(reply.token, reply.data))

    def get_order(
        self,
        pos_id,                 # type: int
        model=None,             # type: Optional[eTree.Element]
        create_order=False,     # type: Optional[bool]
        order_picture=None,     # type: Optional[eTree.Element]
        static_uow=None,        # type: Optional[Any]
    ):
        # type: (...) -> Order

        if not order_picture:
            order_picture = self.get_order_picture(pos_id)

        table_id = static_uow.table_id if static_uow is not None else None
        price_list = static_uow.price_list if static_uow is not None else None
        state = order_picture.get("state") or "" if order_picture else ""
        if state not in ["IN_PROGRESS", "TOTALED"]:
            if not create_order:
                raise OrderNotFoundException()

            order_picture = self._create_new_order(
                pos_id=pos_id,
                model=model,
                price_list=price_list,
                table_id=table_id,
            )

        sale_lines = order_picture.findall(".//SaleLine")

        return Order(
            order_picture=order_picture,
            sale_lines=sale_lines,
            model=model,
            static_uow=static_uow,
        )

    @staticmethod
    def void_order(pos_id, pos_ot):
        # type: (int, OrderTaker) -> None

        pos_ot.voidOrder(pos_id)

    @staticmethod
    def void_line(pos_id, pos_ot, lines_to_void):
        # type: (int, OrderTaker, List) -> None

        if len(lines_to_void) > 0:
            lines_to_void = "|".join([str(x) for x in lines_to_void])
            pos_ot.voidLine(pos_id, lines_to_void)

    @staticmethod
    def clear_discounts(pos_ot, used_discount_ids):
        # type: (OrderTaker, List) -> None

        for discount_id in used_discount_ids:
            pos_ot.clearDiscount(discount_id)

    @staticmethod
    def clear_tenders(pos_id, pos_ot, used_order_tender_ids):
        # type: (int, OrderTaker, List) -> None

        for order_tender_id in used_order_tender_ids:
            pos_ot.clearTenders(pos_id, None, order_tender_id)

    def _create_new_order(
        self,
        pos_id,                 # type: int
        model,                  # type: eTree.Element
        price_list=None,        # type: Optional[str]
        sale_type="EAT_IN",     # type: Optional[str]
        table_id=None,          # type: Optional[str]
    ):
        # type: (...) -> eTree.Element

        # TODO: Review parameters when there is no Order (PriceList and SaleType)
        if price_list is None:
            price_list = "EI"

        pos_ot = sysactions.get_posot(model)
        pos_ot.blkopnotify = True
        order = pos_ot.createOrder(
            posid=pos_id,
            pricelist=price_list,
            saletype=sale_type,
        )

        updated_order = None
        if table_id:
            order_xml = pos_ot.setOrderCustomProperty(
                key="TABLE_ID",
                value=table_id,
                orderid=order.get("orderId"),
            )
            try:
                updated_order = eTree.XML(order_xml)
            except (TypeError, ParseError):
                pass
        pos_ot.blkopnotify = False

        if updated_order is None:
            updated_order = self.get_order_picture(pos_id=pos_id)

        return updated_order

    def get_persisted_benefits_from_order_picture(
        self,
        pos_id,                 # type: int
        order_picture=None,     # type: Optional[eTree]
    ):
        # type: (...) -> Dict

        benefit_list = self.get_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        if not benefit_list:
            return {}

        return benefit_list.get("persistedBenefits", {})



"""
A dict subclass for Python 3 that behaves like Python 2's dict

Example use:

>>> from past.builtins import dict
>>> d1 = dict()    # instead of {} for an empty dict
>>> d2 = dict(key1='value1', key2='value2')

The keys, values and items methods now return lists on Python 3.x and there are
methods for iterkeys, itervalues, iteritems, and viewkeys etc.

>>> for d in (d1, d2):
...     assert isinstance(d.keys(), list)
...     assert isinstance(d.values(), list)
...     assert isinstance(d.items(), list)
"""

import sys

from past.utils import with_metaclass


_builtin_dict = dict
ver = sys.version_info[:2]


class BaseOldDict(type):
    def __instancecheck__(cls, instance):
        return isinstance(instance, _builtin_dict)


class olddict(with_metaclass(BaseOldDict, _builtin_dict)):
    """
    A backport of the Python 3 dict object to Py2
    """
    iterkeys = _builtin_dict.keys
    viewkeys = _builtin_dict.keys

    def keys(self):
        return list(super(olddict, self).keys())

    itervalues = _builtin_dict.values
    viewvalues = _builtin_dict.values

    def values(self):
        return list(super(olddict, self).values())

    iteritems = _builtin_dict.items
    viewitems = _builtin_dict.items

    def items(self):
        return list(super(olddict, self).items())

    def has_key(self, k):
        """
        D.has_key(k) -> True if D has a key k, else False
        """
        return k in self

    # def __new__(cls, *args, **kwargs):
    #     """
    #     dict() -> new empty dictionary
    #     dict(mapping) -> new dictionary initialized from a mapping object's
    #         (key, value) pairs
    #     dict(iterable) -> new dictionary initialized as if via:
    #         d = {}
    #         for k, v in iterable:
    #             d[k] = v
    #     dict(**kwargs) -> new dictionary initialized with the name=value pairs
    #         in the keyword argument list.  For example:  dict(one=1, two=2)

    #     """
    #     
    #     if len(args) == 0:
    #         return super(olddict, cls).__new__(cls)
    #     # Was: elif isinstance(args[0], newbytes):
    #     # We use type() instead of the above because we're redefining
    #     # this to be True for all unicode string subclasses. Warning:
    #     # This may render newstr un-subclassable.
    #     elif type(args[0]) == olddict:
    #         return args[0]
    #     # elif isinstance(args[0], _builtin_dict):
    #     #     value = args[0]
    #     else:
    #         value = args[0]
    #     return super(olddict, cls).__new__(cls, value)
        
    def __native__(self):
        """
        Hook for the past.utils.native() function
        """
        return super(oldbytes, self)


__all__ = ['olddict']




# encoding: utf-8

from typing import Any, Tuple

from justificationdiscountreport.repository import JustificationDiscountRepository
from reportsdata.justificationdiscountreport import (
    JustificationDiscountReportGenerator,
    JustificationDiscountReportPresenter,
)


class ReportGenerator(object):
    @staticmethod
    def generate(
        mb_context,     # type: Any
        pos_id,         # type: str
        *args           # type: str
    ):
        # type: (...) -> Tuple[JustificationDiscountReportGenerator, JustificationDiscountReportPresenter]

        initial_date = args[4]
        end_date = args[5]
        operator_id = None if args[6] == '0' else args[6]

        repository = JustificationDiscountRepository(
            mb_context=mb_context,
            pos_id=pos_id,

        )
        generator = JustificationDiscountReportGenerator(
            repository=repository,
            initial_date=initial_date,
            end_date=end_date,
            operator_id=operator_id,
            pos_id=pos_id,
        )

        presenter = JustificationDiscountReportPresenter(
            columns=38,
        )

        return generator, presenter



# -*- coding: utf-8 -*-
import json

from threading import Lock
from typing import (
    Any,
    Dict,
    List,
    Optional,
    Union,
)
from xml.etree import cElementTree as eTree

from posot import OrderTaker
from sysactions import get_model, get_posot
from messagebus import (
    Event,
    MessageBus,
)

from application.service import OrderService, BenefitList
from application.validator import BenefitValidator
from application.model import Configurations
from application.model.benefit import ActionResult, Benefit, OrderBenefitInfo
from application.model.conditions import Restrictions
from application.model.transactions import (
    AddBenefitToApplyUnitOfWork,
    StaticUnitOfWork,
    VerifyAndRemoveBenefitsUnitOfWork,
)
from application.model.exceptions import (
    CannotRemoveItemsInATotaledOrder,
    CircularDependencyToApplyDiscountException,
    NeedToRemoveBenefitsPreviousBenefitsException,
    OrderValidationException,
)
from application.repository import NDiscountRepository
from application.util import (
    DictionaryOperations,
    benefit_is_add_item_type_on_total,
    get_applied_benefits_by_key,
    get_benefit_key,
    get_benefit_key_from_persisted_benefits,
    get_benefit_to_apply_from_persisted_benefits,
    get_dict_data_remove,
    get_error_message,
    get_list_data_remove,
    get_new_benefit,
    is_order_taker,
    remove_dict_properties_benefits,
    remove_list_properties_benefits,
    split_benefit_key,
    sum_discounts,
)

from ._ActionService import ActionService


class BenefitService(object):
    def __init__(self, configs, repository, message_bus, order_service):
        # type: (Configurations, NDiscountRepository, MessageBus, OrderService) -> None

        self.configs = configs
        self.repository = repository
        self.message_bus = message_bus
        self.order_service = order_service
        self.logger = configs.logger

        self.product_groups = self.repository.get_all_product_groups()
        self.action_service = ActionService(
            order_service=order_service,
            product_groups=self.product_groups,
            message_bus=self.message_bus,
            logger=self.logger,
        )
        self.automatic_benefits = self._get_benefit_types()

        # TODO: Check if we really need a Lock
        self.discount_lock = {}

    def get_static_unit_of_work(
        self,
        pos_id,                 # type: int
        order_picture=None,     # type: Optional[eTree.Element]
        model=None,             # type: Optional[eTree.Element]
        table_id=None,          # type: Optional[str]
        fill_table_info=False,  # type: Optional[bool]
    ):
        # type: (...) -> StaticUnitOfWork

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id=pos_id)

        if fill_table_info and table_id is None:
            table_id = order_picture.find(".//CustomOrderProperties/OrderProperty[@key='{}']".format("TABLE_ID"))
            if table_id is not None:
                table_id = table_id.get("value") or None

        static_uow = StaticUnitOfWork(
            pos_id=pos_id,
            order_picture=order_picture,
            model=model,
            table_id=table_id,
        )
        return static_uow

    def get_unit_of_work(
        self,
        pos_id,                 # type: int
        raw_benefit,            # type: Dict
        static_uow=None,        # type: StaticUnitOfWork
        order_picture=None,     # type: Optional[eTree.Element]
        benefit_applier=None,   # type: Optional[str]
        user_auth_level=False,  # type: Optional[bool]
    ):
        # type: (...) -> AddBenefitToApplyUnitOfWork

        benefit = Benefit(raw_benefit=raw_benefit)
        uow = AddBenefitToApplyUnitOfWork(
            logger=self.logger,
            pos_id=pos_id,
            benefit=benefit,
            benefit_applier=benefit_applier,
            order_picture=order_picture,
            order_service=self.order_service,
            repository=self.repository,
            product_groups=self.product_groups,
            static_uow=static_uow,
            user_auth_level=user_auth_level,
        )

        return uow

    def apply_automatic_benefits_on_table(
            self,
            pos_id,         # type: int
            table_id,       # type: str
            is_tab=False    # type: Optional[bool]
    ):
        # type: (...) -> Optional[str]

        if pos_id not in self.discount_lock:
            self.discount_lock[pos_id] = Lock()
        elif self.discount_lock[pos_id].locked():
            return

        with self.discount_lock[pos_id]:
            try:
                if len(self.automatic_benefits) > 0:
                    discount_type = "ON_TAB_OPEN" if is_tab else "ON_TABLE_OPEN"
                    if self.automatic_benefits.get(discount_type) is not None:
                        return self._apply_open_table_benefits(
                            pos_id=pos_id,
                            table_id=table_id,
                            discount_type=discount_type,
                        )
                    else:
                        return

            except Exception as _:
                self.logger.exception("Error applying automatic discounts on POS {}".format(str(pos_id)))
                raise

    def apply_automatic_benefits_on_tab(
            self,
            pos_id,     # type: int
            table_id    # type: str
    ):
        # type: (...) -> Optional[str]

        if pos_id not in self.discount_lock:
            self.discount_lock[pos_id] = Lock()
        elif self.discount_lock[pos_id].locked():
            return

        with self.discount_lock[pos_id]:
            try:
                if len(self.automatic_benefits) > 0:
                    if self.automatic_benefits.get("ON_TAB_OPEN") is not None:
                        return self._apply_tab_benefits(
                            pos_id=pos_id,
                            table_id=table_id
                        )
                    else:
                        return

            except Exception as _:
                self.logger.exception("Error applying automatic discounts on POS {}".format(str(pos_id)))
                raise

    def _get_benefits_to_apply_on_total_by_priority(
        self,
        pos_id,              # type: int
        order_picture=None,  # type: Optional[eTree.Element]
    ):
        # type: (...) -> list

        benefits_to_apply = self._get_benefits_to_apply_on_total(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        benefits_to_apply = sorted(
            benefits_to_apply,
            key=lambda d: (d.get("priority"), not d.get("is_loyalty_benefit", False))
        )

        return benefits_to_apply

    def _get_benefits_to_apply_on_total(
        self,
        pos_id,              # type: int
        order_picture=None,  # type: Optional[eTree.Element]
    ):
        # type: (...) -> list

        order = self.order_service.get_order(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        all_loyalty_benefits = json.loads(s=order.custom_properties.get("LOYALTY_CUSTOMER_BENEFITS", "[]"))
        selected_benefits_ids = json.loads(s=order.custom_properties.get("LOYALTY_COUPONS_TO_APPLY_ON_TOTAL", "[]"))
        benefits_to_apply = self.automatic_benefits.get("ON_TOTAL", [])[:]

        benefits_from_order = self.order_service.get_applied_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        applied_benefits_ids = [b["benefit"]["id"] for b in benefits_from_order] if benefits_from_order else []

        if not selected_benefits_ids:
            return benefits_to_apply

        for benefit in all_loyalty_benefits:
            if benefit.get("id") in selected_benefits_ids and benefit.get("id") not in applied_benefits_ids:
                benefit["is_loyalty_benefit"] = True
                benefits_to_apply.append(benefit)

        return benefits_to_apply

    def _apply_automatic_benefits_on_sale(
        self,
        pos_id,         # type: int
        order_picture,  # type: eTree.Element
        static_uow,     # type: StaticUnitOfWork
    ):
        # type: (...) -> eTree.Element

        order_state = order_picture.get("state")
        if order_state != "IN_PROGRESS":
            return order_picture

        benefits_on_sale = self.automatic_benefits.get("ON_SALE") or []
        for benefit in benefits_on_sale:
            updated_order = self._apply_automatic_benefit(
                pos_id=pos_id,
                benefit=benefit,
                static_uow=static_uow,
                order_picture=order_picture,
            )
            order_picture = updated_order if updated_order is not None else order_picture

        return order_picture

    def _apply_automatic_benefits_on_total(
        self,
        pos_id,         # type: int
        order_picture,  # type: eTree.Element
        static_uow,     # type: StaticUnitOfWork
    ):
        # type: (...) -> eTree.Element

        benefits_to_apply = self._get_benefits_to_apply_on_total_by_priority(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        persisted_benefits = self.order_service.get_persisted_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        for benefit in benefits_to_apply:
            benefit_key = get_benefit_key_from_persisted_benefits(
                benefit_id=benefit.get("id", ""),
                persisted_benefits=persisted_benefits,
            )
            benefit = get_benefit_to_apply_from_persisted_benefits(
                persisted_benefits=persisted_benefits,
                benefit_key=benefit_key,
                default_benefit=benefit,
            )
            updated_order = self._apply_automatic_benefit(
                pos_id=pos_id,
                benefit=benefit,
                static_uow=static_uow,
                order_picture=order_picture,
                on_total=True,
                persisted_benefit_key=benefit_key,
            )
            order_picture = updated_order if updated_order is not None else order_picture

        return order_picture

    def apply_automatic_benefits_on_order(
        self,
        pos_id,             # type: int
        static_uow=None,    # type: Optional[StaticUnitOfWork]
        on_total=False,     # type: bool
    ):
        # type: (...) -> Optional[eTree.Element]

        if pos_id not in self.discount_lock:
            self.discount_lock[pos_id] = Lock()
        elif self.discount_lock[pos_id].locked():
            return

        with self.discount_lock[pos_id]:
            try:
                if static_uow is None:
                    static_uow = self.get_static_unit_of_work(pos_id=pos_id)

                static_uow.pos_ot.blkopnotify = True
                order_picture = static_uow.order_picture

                if on_total:
                    order_picture = self._apply_automatic_benefits_on_total(
                        pos_id=pos_id,
                        order_picture=order_picture,
                        static_uow=static_uow,
                    )
                else:
                    order_picture = self._apply_automatic_benefits_on_sale(
                        pos_id=pos_id,
                        order_picture=order_picture,
                        static_uow=static_uow,
                    )

            except Exception as _:
                self.logger.exception("Error applying automatic discounts on POS {}".format(str(pos_id)))
                raise

        return order_picture

    def apply_persisted_benefits(
        self,
        pos_id,             # type: int
        static_uow=None,    # type: Optional[StaticUnitOfWork]
    ):
        # type: (...) -> None

        if static_uow is None:
            static_uow = self.get_static_unit_of_work(pos_id=pos_id)

        model = static_uow.model
        if is_order_taker(model=model):
            self.logger.info("The current model is order taker. Ignoring benefit apply for POS: {}".format(pos_id))
            return

        order_picture = static_uow.order_picture
        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        applied_benefits_keys = [b["benefit_key"] for b in applied_benefits]
        benefit_list = self.order_service.get_benefits_from_order_picture(pos_id, order_picture)

        persisted_benefits = benefit_list.get("persistedBenefits", {})
        if not persisted_benefits:
            return

        try:
            pending_benefits = persisted_benefits.get("pending_benefits", {})
            if not pending_benefits:
                return
            for applied_time, benefit_id in persisted_benefits.get("pending_benefits_ids", []):
                benefit_key = "{}_{}".format(benefit_id, applied_time)
                pending_benefit = pending_benefits.get(benefit_key, {})
                is_loyalty_benefit = pending_benefit.get("is_loyalty_benefit")
                if not pending_benefit or benefit_key in applied_benefits_keys or is_loyalty_benefit:
                    return

                uow = self.get_unit_of_work(
                    pos_id=pos_id,
                    raw_benefit=pending_benefit,
                    order_picture=static_uow.order_picture,
                    static_uow=static_uow,
                )
                self.fix_and_apply_benefit(
                    uow=uow,
                    is_manual=True,
                    persisted_key_to_remove=benefit_key,
                )
        except OrderValidationException as _:
            pass
        except Exception as _: # noqa
            exception_message = "Error apply automatic benefit for order {} on POS {}"
            self.logger.exception(exception_message.format(order_picture.get("orderId") or "", str(pos_id)))

    def apply_pending_benefits(
        self,
        pos_id,                             # type: int
        static_uow=None,                    # type: Optional[StaticUnitOfWork]
        previous_error_occurred=False,      # type: Optional[bool]
    ):
        # type: (...) -> None

        if static_uow is None:
            static_uow = self.get_static_unit_of_work(
                pos_id=pos_id
            )

        model = static_uow.model
        if is_order_taker(model=model):
            self.logger.info("The current model is order taker. Ignoring benefit apply for POS: {}".format(pos_id))
            return

        order_picture = static_uow.order_picture
        order_id = order_picture.get("orderId")
        order_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)

        if not order_benefits:
            self.logger.info("No benefits to orderId: {}; posId: {}".format(order_id, pos_id))
            return

        persisted_benefits = self.order_service.get_persisted_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        self.logger.info("Trying to apply benefits to orderId: {}; posId: {}".format(order_id, pos_id))
        for order_benefit_json in order_benefits:
            benefit_order_info = OrderBenefitInfo.from_json(order_benefit_json)
            if benefit_order_info.benefit_applied:
                continue

            self._apply_pending_benefit(
                pos_id=pos_id,
                benefit_order_info=benefit_order_info,
                persisted_benefits=persisted_benefits,
                static_uow=static_uow,
                previous_error_occurred=previous_error_occurred,
            )

    def _send_discount_applied_event(
        self,
        pos_id,    # type: str
        order_id,  # type: str
    ):
        # type: (...) -> None

        try:
            event_data = {
                "posId": pos_id,
                "orderId": order_id,
            }
            event = Event(
                subject="DISCOUNT_APPLIED",
                data=json.dumps(event_data),
            )
            self.message_bus.publish_event(event=event)
            self.logger.info("DISCOUNT_APPLIED event sent successfully! order_id={0}".format(order_id))
        except Exception as _: # noqa
            self.logger.exception("Error sending DISCOUNT_APPLIED event")

    def fix_and_apply_benefit(
        self,
        uow,                                # type: AddBenefitToApplyUnitOfWork
        validate_benefits=True,             # type: Optional[bool]
        on_total=False,                     # type: Optional[bool]
        is_manual=False,                    # type: Optional[bool]
        persisted_key_to_remove='',         # type: Optional[str]
        previous_error_occurred=False,      # type: Optional[bool]
    ):
        # type: (...) -> None

        some_error_occurred = False
        try:
            self._apply_benefit(
                uow=uow,
                validate_benefits=validate_benefits,
                on_total=on_total,
                is_manual=is_manual,
                persisted_key_to_remove=persisted_key_to_remove,
            )
        except NeedToRemoveBenefitsPreviousBenefitsException as ex:
            try:
                some_error_occurred = True

                self.remove_benefits(
                    pos_id=uow.pos_id,
                    benefit_keys=ex.benefits_ids,
                    static_uow=uow.static_uow,
                    delete_from_order=False,
                )

                uow.update_order(order_picture=uow.static_uow.order_picture)

                self._apply_benefit(
                    uow=uow,
                    validate_benefits=validate_benefits,
                    on_total=on_total,
                    is_manual=is_manual,
                    persisted_key_to_remove=persisted_key_to_remove,
                )

                if previous_error_occurred:
                    raise CircularDependencyToApplyDiscountException()
            except NeedToRemoveBenefitsPreviousBenefitsException as _:
                raise CircularDependencyToApplyDiscountException()
            finally:
                if not previous_error_occurred:
                    self.apply_pending_benefits(
                        pos_id=uow.pos_id,
                        static_uow=uow.static_uow,
                        previous_error_occurred=some_error_occurred,
                    )

                    uow.update_order(order_picture=uow.static_uow.order_picture)

    def _apply_benefit(
        self,
        uow,                        # type: AddBenefitToApplyUnitOfWork
        validate_benefits,          # type: Optional[bool]
        on_total,                   # type: Optional[bool]
        is_manual,                  # type: Optional[bool]
        persisted_key_to_remove,    # type: Optional[str]
    ):
        # type: (...) -> None

        try:
            message = "Adding benefit to orderId: {}; posId: {}; benefitName: {}"
            self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name))

            order_picture = uow.order.order_picture
            if order_picture is None:
                order_picture = self.order_service.get_order_picture(uow.pos_id)

            applied_benefits = self.order_service.get_applied_benefits_from_order_picture(uow.pos_id, order_picture)
            burnt_items = self.order_service.get_burnt_items(applied_benefits)

            if validate_benefits:
                BenefitValidator.validate_benefits(uow, applied_benefits, burnt_items, self.product_groups)

            benefit_key = get_benefit_key(benefit_id=uow.benefit_id)
            uow.set_benefit_key(benefit_key)
            action_result = ActionResult([], {}, [], {}, True, False)
            already_burnt = False

            for action in uow.benefit.actions:
                action.times_to_apply = self._get_benefit_times_to_apply(uow)
                restrictions = Restrictions(uow.pos_id, order_picture, uow.benefit, self.repository, self.order_service)
                result = self.action_service.apply_action(action, burnt_items, uow, uow.pos_ot, restrictions, on_total)
                if result.benefit_ids_to_remove:
                    raise NeedToRemoveBenefitsPreviousBenefitsException(benefits_ids=result.benefit_ids_to_remove)

                action_result += result
                items_to_burn = result.items_to_burn
                if len(items_to_burn) > 0 and not already_burnt:
                    updated_items_to_burn = DictionaryOperations.unmerge_counting_dict(items_to_burn, burnt_items)
                    action_result.items_to_burn = updated_items_to_burn
                    already_burnt = True

            if len(uow.benefit.actions) == 0 or action_result:
                updated_order = self._save_benefits_list(
                    uow=uow,
                    pos_ot=uow.pos_ot,
                    action_result=action_result,
                    persisted_key_to_remove=persisted_key_to_remove,
                )
                uow.update_order(order_picture=updated_order)

                if is_manual and action_result.benefit_applied:
                    self._send_discount_applied_event(
                        pos_id=str(uow.pos_id),
                        order_id=str(uow.order.id),
                    )

                message = "Benefit applied successfully to orderId: {}; posId: {}; benefitName: {};"
                self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name))
            else:
                message = "Benefit not applied to orderId: {}; posId: {}; benefitName: {};"
                self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name))
        except OrderValidationException as ex:
            message = "Invalid benefit to orderId: {}; posId: {}; benefitName: {}; validation: {}"
            self.logger.info(message.format(uow.order.id, uow.pos_id, uow.benefit.name, ex.message))
            raise

    @staticmethod
    def _get_benefit_times_to_apply(
        uow,    # type: AddBenefitToApplyUnitOfWork
    ):
        # type: (...) -> int

        return uow.benefit.times_to_apply or (uow.benefit_order_info and uow.benefit_order_info.times_to_apply) or 0

    def validate_benefits(
        self,
        pos_id,                 # type: int
        static_uow,             # type: StaticUnitOfWork
        product_groups=None,    # type: Dict
    ):
        # type: (...) -> eTree

        order_picture = static_uow.order_picture
        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        if not applied_benefits:
            return order_picture

        order = self.order_service.get_order(pos_id, order_picture=order_picture)
        uow = VerifyAndRemoveBenefitsUnitOfWork(pos_id, order)

        log_message = "Validating used benefits from orderId: {}; posId: {}"
        self.logger.info(log_message.format(uow.order.id, uow.pos_id))

        burnt_items = DictionaryOperations.merge_counting_dicts([
            applied_benefit.get("burnt_items", {}) for applied_benefit in applied_benefits
        ])

        removed_benefit = False
        for applied_benefit in applied_benefits:
            benefit = Benefit(applied_benefit.get("benefit"))
            benefit_key = applied_benefit.get("benefit_key")
            benefit_burnt_items = applied_benefit.get("burnt_items", {})
            burnt_items_for_benefit = DictionaryOperations.unmerge_counting_dict(burnt_items, benefit_burnt_items)
            valid_benefit = self._validate_benefit(
                benefit_key=benefit_key,
                benefit=benefit,
                uow=uow,
                burnt_items_for_benefit=burnt_items_for_benefit,
                product_groups=product_groups,
                static_uow=static_uow,
            )
            if valid_benefit is False:
                removed_benefit = True

        if removed_benefit:
            order = self.order_service.get_order(pos_id)
            if sum([x.qty for x in order.sale_lines]) == 0:
                self.order_service.void_order(pos_id, uow.pos_ot)

            return order.order_picture

        return order_picture

    def remove_benefits(
        self,
        pos_id,                  # type: int
        benefit_keys,            # type: List
        static_uow=None,         # type: Optional[StaticUnitOfWork]
        delete_from_order=True,  # type: Optional[bool]
    ):
        # type: (...) -> None

        if static_uow is None:
            static_uow = self.get_static_unit_of_work(pos_id=pos_id)
        order_picture = static_uow.order_picture

        order_id = order_picture.get("orderId")
        log_message = "Trying to remove benefits to orderId: {}; posId: {}; benefitIds: {}"
        self.logger.info(log_message.format(order_id, pos_id, benefit_keys))

        if benefit_keys is None:
            self.logger.info("BenefitIds not provided. Removing all order benefits")
            all_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
            benefit_keys = [x.get("benefit_key") for x in all_benefits]

        self.remove_benefit(
            pos_id=pos_id,
            benefit_keys=benefit_keys,
            delete_from_order=delete_from_order,
            static_uow=static_uow,
        )

    def remove_benefit(
        self,                    # type: BenefitService
        pos_id,                  # type: int
        benefit_keys,            # type: List[str]
        delete_from_order=True,  # type: Optional[bool]
        static_uow=None,         # type: Optional[StaticUnitOfWork]
    ):
        # type: (...) -> None

        order_picture = static_uow.order_picture
        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        if not applied_benefits:
            return

        sorted_applied_benefits = sorted(
            applied_benefits,
            key=lambda b: (b.get("benefit_key").split("_")[-1]),
        )

        benefit_keys_to_remove = get_applied_benefits_by_key(
            benefit_keys=benefit_keys,
            applied_benefits=sorted_applied_benefits,
        )
        self.check_if_benefits_are_removable(
            pos_id=str(pos_id),
            benefits_to_remove=benefit_keys_to_remove,
            order_picture=order_picture,
        )
        benefit_keys_to_remove = [b.get("benefit_key") for b in benefit_keys_to_remove]

        benefits_to_keep = []
        keys_remove = []
        added_sale_lines = []
        used_discount_ids = []
        used_tender_ids = []
        item_amounts_remove = {}

        index = self._get_index_of_first_benefit_to_remove(
            benefit_keys_to_remove=benefit_keys_to_remove,
            sorted_applied_benefits=sorted_applied_benefits,
        )
        benefits_to_keep.extend(sorted_applied_benefits[:index])

        for benefit in sorted_applied_benefits[index:]:
            benefit_key = benefit.get("benefit_key")
            used_discount_ids.extend(benefit.get("used_discount_ids", []))
            item_amounts_remove.update(benefit.get("discount_item_values", {}))
            benefit["used_discount_ids"] = []
            benefit["benefit_applied"] = False

            if benefit_key in benefit_keys_to_remove:
                added_sale_lines.extend(benefit.get("added_sale_lines", []))
                used_tender_ids.extend(benefit.get("order_tenders_ids", []))
                benefit["order_tenders_ids"] = []
                if delete_from_order:
                    keys_remove.append(benefit_key)
                else:
                    benefits_to_keep.append(benefit)

                message = "Remove benefits: Removing benefit. BenefitName: {}; DeleteFromOrder: {}"
                self.logger.info(message.format(benefit["benefit"].get("name"), delete_from_order))
            else:
                if self._benefit_must_be_kept(benefit=benefit):
                    benefit["benefit_applied"] = True

                    message = "Remove benefits: Benefit with added items (not removed). BenefitName: {};"
                    self.logger.info(message.format(benefit["benefit"].get("name")))
                else:
                    used_tender_ids.extend(benefit.get("order_tenders_ids", []))
                    benefit["order_tenders_ids"] = []

                    message = "Remove benefits: Changing to pending. BenefitName: {};"
                    self.logger.info(message.format(benefit["benefit"].get("name")))
                benefits_to_keep.append(benefit)

        pos_ot = static_uow.pos_ot
        self.order_service.void_line(pos_id, pos_ot, added_sale_lines)
        self.order_service.clear_discounts(pos_ot, used_discount_ids)
        self.order_service.clear_tenders(pos_id, pos_ot, used_tender_ids)

        updated_order_picture = self._set_benefits_to_apply(
            pos_id=pos_id,
            benefits=benefits_to_keep,
            pos_ot=pos_ot,
            keys_remove=keys_remove,
            item_amounts_remove=item_amounts_remove,
            tenders_remove=used_discount_ids,
            discounts_remove=used_tender_ids,
            order_picture=order_picture,
        )

        static_uow.set_order_picture(order_picture=updated_order_picture)

    def remove_invalid_benefits(
        self,
        pos_id,         # type: int
        static_uow,     # type: StaticUnitOfWork
    ):
        # type: (...) -> eTree.Element

        order_picture = static_uow.order_picture
        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        applied_benefits = self.order_service.get_applied_benefits_from_order_picture(pos_id, order_picture)
        if not applied_benefits:
            return order_picture

        order = self.order_service.get_order(pos_id, order_picture=order_picture)
        benefits_remove = []
        for applied_benefit in applied_benefits:
            benefit_order_info = OrderBenefitInfo.from_json(applied_benefit)
            if not BenefitValidator.validate_added_sale_line_benefit(benefit_order_info, order):
                benefits_remove.append(benefit_order_info.benefit_key)

        if len(benefits_remove) > 0:
            self.remove_benefits(
                pos_id=pos_id,
                benefit_keys=benefits_remove,
                static_uow=static_uow,
            )

        return order_picture

    def _remove_benefit_by_key(
        self,
        pos_id,               # type: int
        benefit_keys,         # type: List[str]
        keep_as_not_applied,  # type: bool
        static_uow=None,      # type: Optional[StaticUnitOfWork]

    ):
        pass

    @staticmethod
    def _get_index_of_first_benefit_to_remove(
        benefit_keys_to_remove,
        sorted_applied_benefits
    ):
        index = 0
        for benefit in sorted_applied_benefits:
            benefit_key = benefit.get("benefit_key")
            must_be_removed = benefit_key in benefit_keys_to_remove
            if not must_be_removed:
                index += 1
            else:
                break
        return index

    @staticmethod
    def _benefit_must_be_kept(
        benefit,
    ):
        if benefit.get("added_sale_lines"):
            return True

        return False

    def check_if_benefits_are_removable(
        self,
        pos_id,                 # type: str
        benefits_to_remove,     # type: List[Dict]
        order_picture=None,     # type: Optional[eTree.Element]
    ):
        # type: (...) -> None

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        state = order_picture.get("state")

        if state == u"TOTALED":
            for benefit in benefits_to_remove:
                benefit_action = benefit.get("benefit")
                actions = benefit_action.get("actions")
                for action in actions.values():
                    if action.get("itemsToAdd"):
                        raise CannotRemoveItemsInATotaledOrder()

    def _save_benefits_list(
        self,
        uow,                            # type: AddBenefitToApplyUnitOfWork
        pos_ot,                         # type: OrderTaker
        action_result,                  # type: Optional[ActionResult]
        persisted_key_to_remove='',     # type: Optional[str]
    ):
        # type: (...) -> None

        order_picture = uow.order.order_picture
        old_benefits = self.order_service.get_applied_benefits_from_order_picture(uow.pos_id, order_picture)
        benefit_key = uow.benefit_key

        if uow.benefit_order_info is None:
            order_state = order_picture.get("state")
            str_data = get_new_benefit(
                uow=uow,
                action_result=action_result,
                old_benefits=old_benefits,
                benefit_key=benefit_key,
                order_state=order_state,
            )
        else:
            str_data = self._update_benefit(
                uow=uow,
                action_result=action_result,
                order_picture=order_picture,
                new_benefit_key=benefit_key,
            )

        discount_ids = action_result.discounts_id if not uow.benefit.tip_included else []
        tenders_ids = action_result.order_tenders
        item_values = action_result.discount_item_values
        benefit_list = self._get_benefit_list_to_add(
            uow=uow,
            data=str_data,
            benefit_key=benefit_key,
            item_values=item_values,
            order_tenders_ids=tenders_ids,
            discount_ids=discount_ids,
            persisted_key_to_remove=persisted_key_to_remove,
        )
        updated_order = self.order_service.save_benefit_list_custom_property(pos_ot, benefit_list)
        return updated_order

    def _validate_benefit(
        self,
        benefit_key,                # type: str
        benefit,                    # type: Benefit
        uow,                        # type: VerifyAndRemoveBenefitsUnitOfWork
        burnt_items_for_benefit,    # type: Dict
        product_groups,             # type: Dict
        static_uow,                 # type: StaticUnitOfWork
    ):
        # type: (...) -> bool

        try:
            BenefitValidator.validate_benefit_order_info(benefit, uow.order, burnt_items_for_benefit, product_groups)
            return True
        except OrderValidationException:
            self.remove_benefit(
                pos_id=uow.pos_id,
                benefit_keys=[benefit_key],
                static_uow=static_uow,
            )
            return False

    def _update_benefit(
        self,
        uow,                # type: AddBenefitToApplyUnitOfWork
        action_result,      # type: ActionResult
        order_picture,      # type: eTree.Element
        new_benefit_key,    # type: str
    ):
        # type: (...) -> str

        old_benefits = self.order_service.get_benefits_from_order_picture(
            pos_id=uow.pos_id,
            order_picture=order_picture,
        )
        benefit_list = json.loads(old_benefits.get("benefitList", "[]"))
        list_benefit = []
        last_benefit_key = uow.benefit_order_info.benefit_key
        for benefit in benefit_list:
            if benefit["benefit_key"] == last_benefit_key:
                benefit["benefit_key"] = new_benefit_key
                benefit["discount_item_values"] = action_result.discount_item_values
                benefit["used_discount_ids"] = action_result.discounts_id
                benefit["order_tenders_ids"] = action_result.order_tenders
                benefit["benefit_applied"] = action_result.benefit_applied
                if not benefit["benefit_total_value"]:
                    benefit["benefit_total_value"] = sum_discounts(
                        promotion_tender=uow.promotion_tenders,
                        discount_item_values=action_result.discount_item_values,
                    )

            list_benefit.append(benefit)
            list_benefit = sorted(
                list_benefit,
                key=lambda b: (b.get("benefit_key").split("_")[-1]),
            )
        return json.dumps(list_benefit)

    def _set_benefits_to_apply(
        self,
        pos_id,                 # type: int
        benefits,               # type: List[str]
        pos_ot,                 # type: OrderTaker
        keys_remove,            # type: List[str]
        item_amounts_remove,    # type: Dict
        tenders_remove,         # type: List[int]
        discounts_remove,       # type: List[int]
        order_picture=None,     # type: Optional[eTree.Element]
    ):
        # type: (...) -> Optional[eTree.Element]

        sorted_benefits = sorted(
            benefits,
            key=lambda b: (b.get("benefit_key").split("_")[-1]),
        )
        data = json.dumps(sorted_benefits)
        benefit_list = self._get_benefit_list_to_remove(
            pos_id=pos_id,
            data=data,
            benefit_keys_to_remove=keys_remove,
            item_values=item_amounts_remove,
            order_tenders_ids=tenders_remove,
            discounts_remove=discounts_remove,
            order_picture=order_picture,
        )
        if not benefits and not benefit_list.persisted_benefits:
            benefit_list = "clear"

        return self.order_service.save_benefit_list_custom_property(pos_ot, benefit_list)

    def _get_benefit_types(self):
        response = {}
        automatic_benefits = self.repository.get_all_automatic_benefits()
        for benefit in automatic_benefits:
            benefit_json = json.loads(benefit)
            benefit = benefit_json.get("benefit")
            event = benefit.get("automatic")

            if event not in response:
                response[event] = []

            response[event].append(benefit)

        return response

    def _apply_automatic_benefit(
        self,
        pos_id,                         # type: int
        benefit,                        # type: Dict
        static_uow,                     # type: StaticUnitOfWork
        order_picture=None,             # type: Optional[eTree.Element]
        on_total=False,                 # type: Optional[bool]
        persisted_benefit_key=None,     # type: Optional[str]
    ):
        # type: (...) -> Optional[eTree.Element]

        order_id = None
        uow = None
        try:
            uow = self.get_unit_of_work(
                pos_id=pos_id,
                raw_benefit=benefit,
                static_uow=static_uow,
                order_picture=order_picture,
            )
            order_id = uow.order.id
            self.fix_and_apply_benefit(
                uow=uow,
                on_total=on_total,
                persisted_key_to_remove=persisted_benefit_key,
            )
            return uow.order.order_picture
        except OrderValidationException as _:
            pass
        except Exception as _:
            self.logger.exception(get_error_message(
                order_id=order_id,
                benefit=uow.benefit,
            ))

    def _apply_pending_benefit(
        self,
        pos_id,                             # type: int
        benefit_order_info,                 # type: OrderBenefitInfo
        persisted_benefits,                 # type: Dict
        static_uow,                         # type: StaticUnitOfWork
        previous_error_occurred=False,      # type: Optional[bool]
    ):
        # type: (...) -> None

        order_id = None
        benefit = None
        try:
            benefit = benefit_order_info.benefit

            benefit_raw = get_benefit_to_apply_from_persisted_benefits(
                persisted_benefits=persisted_benefits,
                benefit_key=benefit_order_info.benefit_key,
                default_benefit=benefit.raw,
            )

            uow = self.get_unit_of_work(
                pos_id=pos_id,
                raw_benefit=benefit_raw,
                static_uow=static_uow,
                order_picture=static_uow.order_picture,
            )
            uow.benefit_order_info = benefit_order_info
            order_id = uow.order.id
            self.fix_and_apply_benefit(
                uow=uow,
                validate_benefits=False,
                persisted_key_to_remove=benefit_order_info.benefit_key or "",
                previous_error_occurred=previous_error_occurred,
            )
        except OrderValidationException as _:
            pass
        except CircularDependencyToApplyDiscountException as _:
            raise
        except Exception as _:
            self.logger.exception(get_error_message(
                order_id=order_id,
                benefit=benefit,
            ))

    def _get_benefit_list_to_add(
        self,
        uow,                        # type: AddBenefitToApplyUnitOfWork
        data,                       # type: str
        benefit_key,                # type: Any[str, List[str]]
        item_values,                # type: Dict
        order_tenders_ids,          # type: List
        discount_ids,               # type: List
        persisted_key_to_remove,    # type: str
    ):
        # type: (...) -> BenefitList

        order_picture = uow.order.order_picture
        benefits_pre_total = self._get_benefits_pre_total(
            pos_id=uow.pos_id,
            benefit_key_to_add=benefit_key,
            order_picture=order_picture,
        )
        persisted_benefits_on_total = self._get_persisted_benefits(
            pos_id=uow.pos_id,
            order_picture=order_picture,
            persisted_key_to_remove=persisted_key_to_remove,
        )
        discount_item_values = self._get_discount_item_values(uow.pos_id, item_values, order_picture)
        discount_excluded = self._get_discount_type_excluded(uow.pos_id, discount_ids, order_picture)
        tender_excluded = self._get_tender_type_excluded(uow.pos_id, order_tenders_ids, order_picture)

        return BenefitList(
            benefit_pre_total_sequence=benefits_pre_total,
            persisted_benefits=persisted_benefits_on_total,
            discount_item_values=discount_item_values,
            discount_id_tip_excluded=discount_excluded,
            tender_id_tip_excluded=tender_excluded,
            benefits=data,
        )

    def _get_benefit_list_to_remove(
        self,
        pos_id,                     # type: int
        data,                       # type: str
        benefit_keys_to_remove,     # type: Any[str, List[str]]
        item_values,                # type: Dict
        order_tenders_ids,          # type: List[int]
        discounts_remove,           # type: List[int]
        order_picture=None,         # type: Optional[eTree.Element]
    ):
        # type: (...) -> BenefitList

        if order_picture is None:
            order_picture = self.order_service.get_order_picture(pos_id)

        benefits_pre_total = self._get_benefits_pre_total(
            pos_id=pos_id,
            benefits_keys_to_remove=benefit_keys_to_remove,
            order_picture=order_picture,
        )
        persisted_benefits_on_total = self._get_persisted_benefits(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        discount_item_values = self._remove_properties(pos_id, item_values, "discountItemValues", order_picture, True)
        discount_excluded = self._remove_properties(pos_id, discounts_remove, "discountIdTipExcluded", order_picture)
        tender_excluded = self._remove_properties(pos_id, order_tenders_ids, "tenderIdTipExcluded", order_picture)

        return BenefitList(
            benefit_pre_total_sequence=benefits_pre_total,
            persisted_benefits=persisted_benefits_on_total,
            discount_item_values=discount_item_values,
            discount_id_tip_excluded=discount_excluded,
            tender_id_tip_excluded=tender_excluded,
            benefits=data,
        )

    def _get_benefits_pre_total(
        self,
        pos_id,                         # type: int
        benefit_key_to_add=None,        # type: Any[Union[str, List[str]]]
        benefits_keys_to_remove=None,   # type: List[str]
        order_picture=None,             # type: Optional[eTree.Element]
    ):
        # type: (...) -> List

        benefits = self.order_service.get_benefits_from_order_picture(pos_id, order_picture)
        if order_picture.get("state") == "TOTALED":
            return benefits.get("benefitPreTotalSequence", [])

        if benefits_keys_to_remove:
            return remove_list_properties_benefits(
                benefits=benefits,
                list_remove=benefits_keys_to_remove,
                property_name="benefitPreTotalSequence",
            )

        pre_total_sequence = benefits.get("benefitPreTotalSequence", [])
        if benefit_key_to_add:
            pre_total_sequence.append(benefit_key_to_add)
        return pre_total_sequence

    def _get_persisted_benefits(
        self,
        pos_id,                         # type: int
        order_picture=None,             # type: Optional[eTree.Element]
        persisted_key_to_remove='',     # type: Optional[str]
    ):
        # type: (...) -> Dict

        benefits = self.order_service.get_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        persisted_benefits = benefits.get("persistedBenefits", {})
        if not persisted_benefits or not persisted_key_to_remove:
            return persisted_benefits

        benefit_id, applied_time = split_benefit_key(benefit_key=persisted_key_to_remove)
        persisted_benefits["pending_benefits_ids"].remove([applied_time, benefit_id])
        persisted_benefits["pending_benefits"].pop(persisted_key_to_remove)

        if not persisted_benefits["pending_benefits_ids"]:
            return {}

        return persisted_benefits

    def _get_discount_item_values(
        self,
        pos_id,                 # type: int
        item_discount_values,   # type: Dict
        order_picture=None,     # type: Optional[eTree.Element]
    ):
        # type: (...) -> Dict

        benefits = self.order_service.get_benefits_from_order_picture(pos_id, order_picture)
        return get_dict_data_remove(
            benefits=benefits,
            dict_data=item_discount_values,
            property_name="discountItemValues",
        )

    def _get_discount_type_excluded(
        self,
        pos_id,                 # type: int
        discount_ids,           # type: List[str]
        order_picture=None,     # type: Optional[eTree.Element]
    ):
        # type: (...) -> List

        benefits = self.order_service.get_benefits_from_order_picture(pos_id, order_picture)
        return get_list_data_remove(
            benefits=benefits,
            list_data=discount_ids,
            property_name="discountIdTipExcluded",
        )

    def _get_tender_type_excluded(
        self,
        pos_id,                 # type: int
        order_tenders_ids,      # type: List[str]
        order_picture=None,     # type: Optional[eTree.Element]
    ):
        # type: (...) -> List

        benefits = self.order_service.get_benefits_from_order_picture(pos_id, order_picture)
        if len(benefits) == 0 or len(order_tenders_ids) == 0:
            return order_tenders_ids

        return get_list_data_remove(
            benefits=benefits,
            list_data=order_tenders_ids,
            property_name="tenderIdTipExcluded",
        )

    def _remove_properties(
        self,
        pos_id,                 # type: int
        data,                   # type: Union[Dict, List]
        property_name,          # type: str
        order_picture,          # type: eTree.Element
        is_dict_data=False,     # type: Optional[bool]
    ):
        # type: (...) -> Union[Dict, List]

        benefits = self.order_service.get_benefits_from_order_picture(pos_id, order_picture)

        if is_dict_data:
            return remove_dict_properties_benefits(
                benefits=benefits,
                dict_remove=data,
                property_name=property_name,
            )
        return remove_list_properties_benefits(
            benefits=benefits,
            list_remove=data,
            property_name=property_name,
        )

    def _apply_open_table_benefits(
        self,
        pos_id,         # type: int
        table_id,       # type: str
        discount_type,  # type: str
    ):
        # type: (...) -> Optional[str]

        static_uow = self.get_static_unit_of_work(
            pos_id=pos_id,
            table_id=table_id,
        )
        static_uow.pos_ot.blkopnotify = True
        order_picture = static_uow.order_picture

        for benefit in self.automatic_benefits[discount_type]:
            updated_order = self._apply_automatic_benefit(
                pos_id=pos_id,
                benefit=benefit,
                static_uow=static_uow,
                order_picture=order_picture,
            )
            order_picture = updated_order if updated_order is not None else order_picture

        order = self.order_service.get_order(
            pos_id=pos_id,
            order_picture=order_picture,
        )

        order_id = order.id if order else ''
        if sum([x.qty for x in order.sale_lines]) == 0:
            model = get_model(pos_id)
            pos_ot = get_posot(model)
            self.order_service.void_order(pos_id, pos_ot)
            return
        return str(order_id)

    def save_persistent_benefits(
        self,
        pos_id,         # type: int
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> eTree.Element

        benefit_list = self.order_service.get_benefits_from_order_picture(
            pos_id=pos_id,
            order_picture=order_picture,
        )
        if not benefit_list:
            return order_picture

        all_benefits = json.loads(benefit_list.get("benefitList", "[]"))
        applied_discounts = [x for x in all_benefits if x['benefit_applied'] is True]
        if not applied_discounts:
            return order_picture

        persisted_benefits = dict(
            pending_benefits_ids=[],
            pending_benefits={},
        )
        for discount in applied_discounts:
            benefit = discount.get("benefit", {})
            persist = benefit.get("persistent", None)
            if not persist:
                continue

            if benefit_is_add_item_type_on_total(benefit=benefit):
                continue

            if persist == "KEEP_VALUE" and "persistentValue" not in benefit.keys():
                benefit["persistentValue"] = discount.get("benefit_total_value", 0)

            benefit_key = discount.get("benefit_key")
            benefit_id, applied_time = split_benefit_key(benefit_key=benefit_key)
            persisted_benefits["pending_benefits_ids"].append([applied_time, benefit_id])
            persisted_benefits["pending_benefits"].update({benefit_key: benefit})

        persisted_benefits["pending_benefits_ids"].sort()

        benefit_list["persistedBenefits"] = persisted_benefits

        if persisted_benefits["pending_benefits_ids"]:
            pos_ot = get_posot(model=get_model(posid=pos_id))
            self.order_service.save_benefit_list_custom_property(
                pos_ot=pos_ot,
                benefit_list=benefit_list,
            )
            return self.order_service.get_order_picture(pos_id=pos_id)

        return order_picture

    @staticmethod
    def send_order_update_event(
        static_uow,     # type: StaticUnitOfWork
    ):
        # type: (...) -> None

        static_uow.pos_ot.blkopnotify = False
        pos_ot = static_uow.pos_ot
        sale_type = static_uow.order_picture.get("saleType")
        pos_ot.updateOrderProperties(
            posid=static_uow.pos_id,
            saletype=sale_type,
        )



from abc import ABCMeta, abstractmethod

from typing import List, Optional  # noqa


class ListBox(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def show_listbox(self, pos_id, options, message='$SELECT_AN_OPTION', title='', def_value='',
                     buttons='$OK|$CANCEL', icon='', timeout=60000, asynch=False):
        # type: (str, List[str], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], Optional[int], Optional[bool]) -> int
        raise NotImplementedError()



from ._CustomExceptions import (
    NotFound,
    BackupAlertsDatabaseNoSourceException,
    RestoreAlertsDatabaseNoSourceException,
    NoResponseError,
    ForbiddenError,
    NotFoundError,
    NotAvailableError,
    InUseError,
    ProcessorError,
)



# -*- coding: utf-8 -*-
# flake8: noqa
import marshal
import sys

PY2 = sys.version_info[0] == 2
PYPY = hasattr(sys, "pypy_translation_info")
_identity = lambda x: x

if not PY2:
    unichr = chr
    range_type = range
    text_type = str
    string_types = (str,)
    integer_types = (int,)

    iterkeys = lambda d: iter(d.keys())
    itervalues = lambda d: iter(d.values())
    iteritems = lambda d: iter(d.items())

    import pickle
    from io import BytesIO, StringIO

    NativeStringIO = StringIO

    def reraise(tp, value, tb=None):
        if value.__traceback__ is not tb:
            raise value.with_traceback(tb)
        raise value

    ifilter = filter
    imap = map
    izip = zip
    intern = sys.intern

    implements_iterator = _identity
    implements_to_string = _identity
    encode_filename = _identity

    marshal_dump = marshal.dump
    marshal_load = marshal.load

else:
    unichr = unichr
    text_type = unicode
    range_type = xrange
    string_types = (str, unicode)
    integer_types = (int, long)

    iterkeys = lambda d: d.iterkeys()
    itervalues = lambda d: d.itervalues()
    iteritems = lambda d: d.iteritems()

    import cPickle as pickle
    from cStringIO import StringIO as BytesIO, StringIO

    NativeStringIO = BytesIO

    exec("def reraise(tp, value, tb=None):\n raise tp, value, tb")

    from itertools import imap, izip, ifilter

    intern = intern

    def implements_iterator(cls):
        cls.next = cls.__next__
        del cls.__next__
        return cls

    def implements_to_string(cls):
        cls.__unicode__ = cls.__str__
        cls.__str__ = lambda x: x.__unicode__().encode("utf-8")
        return cls

    def encode_filename(filename):
        if isinstance(filename, unicode):
            return filename.encode("utf-8")
        return filename

    def marshal_dump(code, f):
        if isinstance(f, file):
            marshal.dump(code, f)
        else:
            f.write(marshal.dumps(code))

    def marshal_load(f):
        if isinstance(f, file):
            return marshal.load(f)
        return marshal.loads(f.read())


def with_metaclass(meta, *bases):
    """Create a base class with a metaclass."""
    # This requires a bit of explanation: the basic idea is to make a
    # dummy metaclass for one level of class instantiation that replaces
    # itself with the actual metaclass.
    class metaclass(type):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)

    return type.__new__(metaclass, "temporary_class", (), {})


try:
    from urllib.parse import quote_from_bytes as url_quote
except ImportError:
    from urllib import quote as url_quote


try:
    from collections import abc
except ImportError:
    import collections as abc


try:
    from os import fspath
except ImportError:
    try:
        from pathlib import PurePath
    except ImportError:
        PurePath = None

    def fspath(path):
        if hasattr(path, "__fspath__"):
            return path.__fspath__()

        # Python 3.5 doesn't have __fspath__ yet, use str.
        if PurePath is not None and isinstance(path, PurePath):
            return str(path)

        return path



class StatusResponse(object):
    def __init__(
        self,
        status,     # type: bool
        message,    # type: str
    ):
        # type: (...) -> None

        self.status = status
        self.message = message



from abc import ABCMeta, abstractmethod
from typing import Any


class MessageProcessorCallback(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def process_begun(self, event_name, request_id, data):
        # type: (str, str, str) -> None
        raise NotImplementedError()

    @abstractmethod
    def input_parsed(self, event_name, request_id, input_obj):
        # type: (str, str, Any) -> None
        raise NotImplementedError()

    @abstractmethod
    def business_called(self, event_name, request_id, input_obj, output_obj):
        # type: (str, str, Any, Any) -> None
        raise NotImplementedError()

    @abstractmethod
    def process_finished(self, event_name, request_id, input_obj, output_obj):
        # type: (str, str, Any, BaseException) -> None
        raise NotImplementedError()

    @abstractmethod
    def parse_exception(self, event_name, request_id, data, exception):
        # type: (str, str, str, BaseException) -> None
        raise NotImplementedError()

    @abstractmethod
    def unhandled_exception(self, event_name, request_id, input_obj, exception):
        # type: (str, str, Any, BaseException) -> None
        raise NotImplementedError()



from typing import List, Dict


class BenefitList(object):
    def __init__(
        self,
        benefit_pre_total_sequence,     # type: List
        persisted_benefits,             # type: List
        discount_item_values,           # type: Dict
        discount_id_tip_excluded,       # type: List
        tender_id_tip_excluded,         # type: List
        benefits,                       # type: str
    ):
        # type: (...) -> None

        self.benefit_pre_total_sequence = benefit_pre_total_sequence
        self.persisted_benefits = persisted_benefits
        self.discount_item_values = discount_item_values
        self.discount_id_tip_excluded = discount_id_tip_excluded
        self.tender_id_tip_excluded = tender_id_tip_excluded
        self.benefit_list = benefits



from typing import (
    Optional,
)
from xml.etree import cElementTree as eTree

from msgbus import (
    FM_PARAM,
    TK_CMP_MESSAGE,
    TK_SYS_NAK,
)
from messagebus import (
    Message,
    MessageBus,
)


class TablePictureRepository(object):
    def __init__(
        self,
        message_bus,      # type: MessageBus
    ):
        # type: (...) -> None

        self.message_bus = message_bus

    def get_table_info(
        self,
        pos_id,      # type: int
        table_id,    # type: int
    ):
        # type: (...) -> eTree

        if not table_id:
            return None

        params = ("TABLE_PICTURE", pos_id, table_id)
        req_buf = "\0".join([str(p) for p in params])
        message = Message(
            token=TK_CMP_MESSAGE,
            data_type=FM_PARAM,
            data=req_buf,
        )
        msg = self.message_bus.send_message(
            component='TABLEMGR',
            message=message,
        )

        if msg.token != TK_SYS_NAK:
            return eTree.XML(msg.data)

        return None



import json
from json import dumps
from xml.etree.ElementTree import Element

from typing import Optional, Dict

from ._ServicePeriod import ServicePeriod


class SaleLine(object):
    def __init__(self, obj):
        # type: (Element) -> None

        self.qty = float(obj.get("multipliedQty") or 0)
        self.context = obj.get("itemId")
        self.item_id = obj.get("partCode")
        self.price_key = int(obj.get("priceKey")) if obj.get("priceKey") else None
        self.item_price = float(obj.get("itemPrice") or 0)
        unit_price = float(obj.get("overwrittenUnitPrice", obj.get("unitPrice") or 0))
        self.unit_price = float(obj.get("addedUnitPrice") or 0) if unit_price == 0 else unit_price
        self.unit_discount = float(obj.get("itemDiscount") or 0) / (self.qty or 1)
        self.current_unit_price = self.unit_price - self.unit_discount
        self.level = obj.get("level")
        self.line_number = obj.get("lineNumber")
        self.key = "{}.{}.{}.{}".format(self.line_number, self.level, self.context, self.item_id)
        self.product_name = obj.get("productName") or ""
        self.item_type = obj.get("itemType")

        self._get_custom_properties_info(obj)

    def _get_custom_properties_info(self, obj):
        self.seat = None
        self.service_period = None
        self.sale_type = None
        self.brand = None
        custom_properties = self.get_custom_properties(obj)
        if custom_properties:
            self.seat = custom_properties.get("seat")
            self.service_period = ServicePeriod(custom_properties.get("servicePeriod"))
            self.sale_type = custom_properties.get("saleType")
            self.brand = custom_properties.get("brand")

    @staticmethod
    def get_custom_properties(obj):
        # type: (Element) -> Optional[Dict]

        try:
            custom_properties_json = obj.get("customProperties")
            if not custom_properties_json:
                return None

            custom_properties = json.loads(custom_properties_json)
            return custom_properties
        except Exception as _:
            return None

    def to_json(self):
        return dumps(self, default=lambda o: o.__dict__ if hasattr(o, '__dict__') else str(o), sort_keys=True)



from unittest import TestCase
import simplejson as json

def default_iterable(obj):
    return list(obj)

class TestCheckCircular(TestCase):
    def test_circular_dict(self):
        dct = {}
        dct['a'] = dct
        self.assertRaises(ValueError, json.dumps, dct)

    def test_circular_list(self):
        lst = []
        lst.append(lst)
        self.assertRaises(ValueError, json.dumps, lst)

    def test_circular_composite(self):
        dct2 = {}
        dct2['a'] = []
        dct2['a'].append(dct2)
        self.assertRaises(ValueError, json.dumps, dct2)

    def test_circular_default(self):
        json.dumps([set()], default=default_iterable)
        self.assertRaises(TypeError, json.dumps, [set()])

    def test_circular_off_default(self):
        json.dumps([set()], default=default_iterable, check_circular=False)
        self.assertRaises(TypeError, json.dumps, [set()], check_circular=False)



from report.command import Command  # noqa
from typing import List, Optional  # noqa


class Part(object):
    def __init__(
        self,
        text=None,          # type: Optional[str]
        commands=None,      # type: Optional[List[Command]]
        encoding='ASCII',   # type: Optional[str]
    ):
        # type: (...) -> None

        self.text = text
        self.commands = commands
        self.encoding = encoding



from messagebus import TokenCreator, TokenPriority


class LoyaltyToken(object):
    LOYALTY_MESSAGE_GROUP = "111"

    @staticmethod
    def create_token(
        number,  # type: str
    ):
        return TokenCreator.create_token(
            priority=TokenPriority.low,
            message_group=LoyaltyToken.LOYALTY_MESSAGE_GROUP,
            number=number
        )



import unittest
from simplejson.compat import StringIO

import simplejson as json

def iter_dumps(obj, **kw):
    return ''.join(json.JSONEncoder(**kw).iterencode(obj))

def sio_dump(obj, **kw):
    sio = StringIO()
    json.dumps(obj, **kw)
    return sio.getvalue()

class TestIterable(unittest.TestCase):
    def test_iterable(self):
        for l in ([], [1], [1, 2], [1, 2, 3]):
            for opts in [{}, {'indent': 2}]:
                for dumps in (json.dumps, iter_dumps, sio_dump):
                    expect = dumps(l, **opts)
                    default_expect = dumps(sum(l), **opts)
                    # Default is False
                    self.assertRaises(TypeError, dumps, iter(l), **opts)
                    self.assertRaises(TypeError, dumps, iter(l), iterable_as_array=False, **opts)
                    self.assertEqual(expect, dumps(iter(l), iterable_as_array=True, **opts))
                    # Ensure that the "default" gets called
                    self.assertEqual(default_expect, dumps(iter(l), default=sum, **opts))
                    self.assertEqual(default_expect, dumps(iter(l), iterable_as_array=False, default=sum, **opts))
                    # Ensure that the "default" does not get called
                    self.assertEqual(
                        expect,
                        dumps(iter(l), iterable_as_array=True, default=sum, **opts))



from typing import List

from ._TipReportDetailLineDto import TipReportDetailLineDto


class TipReportBodyDto(object):
    def __init__(self):
        # type: () -> None
        self.__detail_line_list = []
        self.__total_tables_count = 0
        self.__total_order_count = 0
        self.__total_sold_amount = 0.0
        self.__total_tip_amount = 0.0

    @property
    def total_tables_count(self):
        # type: () -> str
        return str(self.__total_tables_count)

    @property
    def total_order_count(self):
        # type: () -> str
        return str(self.__total_order_count)

    @property
    def total_sold_amount(self):
        # type: () -> str
        return str(self.__total_sold_amount)

    @property
    def total_tip_amount(self):
        # type: () -> str
        return str(self.__total_tip_amount)

    @property
    def detail_line_list(self):
        # type: () -> List[TipReportDetailLineDto]
        return self.__detail_line_list

    def add_detail_line(self, tip_report_detail_line_dto):
        # type: (TipReportDetailLineDto) -> None

        self.__total_tables_count += int(tip_report_detail_line_dto.table_count)
        self.__total_order_count += int(tip_report_detail_line_dto.order_count)
        self.__total_sold_amount += round(float(tip_report_detail_line_dto.total_sold_amount), 2)
        self.__total_tip_amount += round(float(tip_report_detail_line_dto.tip_amount), 2)

        self.detail_line_list.append(tip_report_detail_line_dto)



from ._FiscalServiceModel import FiscalServiceModel
from ._OrderServiceModel import OrderServiceModel
from ._ApiServiceModel import ApiServiceModel
from ._CouponServiceModel import CouponServiceModel
from ._ImagesServiceModel import ImagesServiceModel
from ._PaymentServiceModel import PaymentServiceModel
from ._ProductDescription import ProductInfoServiceModel
from ._LoyaltyServiceModel import LoyaltyServiceModel
from ._ElectronicPaymentServiceModel import ElectronicPaymentServiceModel
from ._NavigationServiceModel import NavigationServiceModel



from typing import Optional


class CatalogVersionDto(object):
    def __init__(
        self,
        menu_id,        # type: Optional[str]
        menu_version,   # type: Optional[str]
        menu_name,      # type: Optional[str]
    ):
        # type: (...) -> None

        self.menu_id = menu_id
        self.menu_version = menu_version
        self.menu_name = menu_name



# -*- coding: utf-8 -*-
from xml.etree import cElementTree as eTree
from xml.etree.ElementTree import Element

import pyscripts
import sysactions
from msgbus import (
    TK_SYS_ACK,
    TK_SLCTRL_OMGR_ORDERPICT,
    FM_PARAM,
)
from pos_model import TableStatus
from typing import (
    Optional,
    Union,
    Any,
)

from application.repository import AccountRepository
from actions.login.common_lock import (
    acquire_user_lock,
    release_user_lock,
)
from actions.util import get_operator_info, get_user_level
from actions.void.do_void_order import do_void_order

from utilfunctions import wait_model_update
from ..functions.do_store_order import doStoreOrder
from ..functions.do_transfer import do_transfer, get_justification
from .. import logger, mb_context
from .utils import (
    close_user,
)

USECS_PER_SEC = 1000000


def close_order_from_another_pos(
    model,                                  # type: eTree.XML
    pos_id,                                 # type: str
    user_id,                                # type: str
    pos_to_close_user,                      # type: str
    is_quick_service,                       # type: bool
    table_service=None,                     # type: Any
    show_confirmation_message=True,         # type: Optional[bool]
    must_store_quick_servie_order=False,    # type: Optional[bool]
):
    # type: (...) -> bool

    if sysactions.has_current_order(model=model):
        order = sysactions.get_current_order(model=model)
        if is_quick_service:
            can_close_order = _can_close_quick_service_order(
                order=order,
                pos_id=pos_id,
                pos_to_close_user=pos_to_close_user,
                show_confirmation_message=show_confirmation_message,
                must_store_quick_servie_order=must_store_quick_servie_order,
            )
            if not can_close_order:
                return False
        else:
            can_close_order = _can_close_table_service_order(
                order=order,
                pos_id=pos_id,
                user_id=user_id,
                pos_to_close_user=pos_to_close_user,
                table_service=table_service,
                show_confirmation_message=show_confirmation_message,
            )
            if not can_close_order:
                return False

    _send_table_modified_event(
        pos_to_close_user=pos_to_close_user,
    )

    return True


def _can_close_quick_service_order(
    order,                                  # type: eTree.XML
    pos_id,                                 # type: str
    pos_to_close_user,                      # type: str
    show_confirmation_message=True,         # type: Optional[bool]
    must_store_quick_servie_order=False,    # type: Optional[bool]
):
    # type: (...) -> bool

    if order.get("state") in ("IN_PROGRESS", "TOTALED"):
        if must_store_quick_servie_order and order_was_already_saved(order=order):
            doStoreOrder(
                pos_id=pos_to_close_user,
                pos_id_to_show_message=pos_id,
                show_popup=False,
            )
            return True

        cancel_operation = _show_message_to_void_order(
            pos_id=pos_id,
            pos_to_close_user=pos_to_close_user,
            show_confirmation_message=show_confirmation_message,
        )
        if cancel_operation:
            return False

    return True


def _can_close_table_service_order(
    order,                                              # type: eTree.XML
    pos_id,                                             # type: str
    user_id,                                            # type: str
    pos_to_close_user,                                  # type: str
    table_service=None,                                 # type: Any
    show_confirmation_message=True,                     # type: Optional[bool]
):
    # type: (...) -> bool

    if order.get("state") == "IN_PROGRESS" and not order_was_already_saved(order=order):
        cancel_operation = _show_message_to_void_order(
            pos_id=pos_id,
            pos_to_close_user=pos_to_close_user,
            show_confirmation_message=show_confirmation_message,
        )
        if cancel_operation:
            return False
    else:
        if table_service is not None:
            all_tables = table_service().list_tables(pos_id=pos_to_close_user)
            for table in all_tables:
                _validate_if_table_has_order_opened(
                    pos_id=pos_id,
                    pos_to_close_user=pos_to_close_user,
                    table=table,
                    table_service=table_service,
                    user_id=user_id,
                )

    return True


def _send_table_modified_event(
    pos_to_close_user,  # type: str
):
    # type: (...) -> None

    create_table_modified_event(
        pos_id=pos_to_close_user,
        table_picture='0',
    )

    wait_model_update(
        pos_id=pos_to_close_user,
        model_update_validation=sysactions.has_current_order,
    )


def order_was_already_saved(
    order,  # type: eTree.XML
):
    # type: (...) -> bool

    state_history = order.findall(".//StateHistory/State")
    for state in state_history:
        if state.get("state") == "STORED":
            return True

    return False


def _validate_if_table_has_order_opened(
    pos_id,             # type: str
    pos_to_close_user,  # type: str
    table,              # type: Any
    table_service,      # type: Any
    user_id,            # type: str
):
    # type: (...) -> None

    if user_id:
        table_picture = eTree.XML(table_service().get_table_picture(
            pos_id=pos_to_close_user,
            table_id=table.id,
        ))

        is_same_user = table_picture.get("userId") == user_id
        valid_status = table.status in (TableStatus.TOTALED, TableStatus.IN_PROGRESS)
        if is_same_user and valid_status and table.current_pos_id is not None:
            doStoreOrder(
                pos_id=table.current_pos_id,
                pos_id_to_show_message=pos_id,
                show_popup=False,
            )


def _show_message_to_void_order(
    pos_id,                     # type: str
    pos_to_close_user,          # type: str
    show_confirmation_message,  # type: bool
):
    # type: (...) -> bool

    confirmation = 0
    if show_confirmation_message:
        confirmation = sysactions.show_messagebox(
            posid=pos_id,
            message="$CLOSE_USER_VOID_ORDER",
            timeout=720000,
            buttons="$OK|$CANCEL",
        )

    if confirmation == 0:
        do_void_order(
            pos_id=pos_to_close_user,
            received_void_reason="CANCELLATION",
            show_cancel_message=False,
        )
        return False

    return True


def get_order_picture(
    pos_id,         # type: str
    order_id=None,  # type: str
):
    # type: (Union[str, int], Optional[str]) -> Optional[Element]

    data = "{}\0{}".format(str(pos_id), str(order_id or ''))
    reply = sysactions.send_message(
        dest_name="ORDERMGR{}".format(str(pos_id)),
        token=TK_SLCTRL_OMGR_ORDERPICT,
        format=FM_PARAM,
        data=data,
    )

    if reply.token == TK_SYS_ACK:
        try:
            split_reply = reply.data.split("\0")
            if not split_reply[2]:
                return None
            order_picture = eTree.XML(split_reply[2])
            if order_id:
                order_picture = order_picture.findall('.//Order')[0]
            return order_picture
        except AttributeError:
            return None

    raise Exception("Error getting OrderPicture: {} - {}".format(reply.token, reply.data))


def create_table_modified_event(
    pos_id,         # type: str
    table_picture,  # type: str
):
    # type: (...) -> None

    event_xml = """<Event subject="POS{}" type="TABLE_MODIFIED">
                           <TABLE_MODIFIED>
                           {}
                       </TABLE_MODIFIED>
                   </Event>""".format(pos_id, table_picture)
    mbcontext = pyscripts.mbcontext
    mbcontext.MB_EasyEvtSend(
        subject="POS{}".format(int(pos_id)),
        type="TableSelected",
        xml=event_xml,
    )


def common_close_user(
    pos_id,                     # type: str
    model,                      # type: eTree.XML
    user_id,                    # type: str
    authorizer=None,            # type: Optional[str]
    pos_to_close_user=None,     # type: Optional[str]
    table_service=None,         # type: Optional[Any]
):
    # type: (...) -> bool

    if not acquire_user_lock(user_id=user_id):
        sysactions.show_messagebox(
            posid=pos_id,
            message="$OPERATOR_IS_PERFORMING_AN_OPEN_OR_CLOSE_ACTION",
            icon="warning",
        )
        raise sysactions.StopAction()

    try:
        return _inner_common_close_user(
            pos_id=pos_id,
            model=model,
            user_id=user_id,
            authorizer=authorizer,
            pos_to_close_user=pos_to_close_user,
            table_service=table_service,
        )
    finally:
        release_user_lock(user_id=user_id)


def _inner_common_close_user(
    pos_id,                     # type: str
    model,                      # type: eTree.XML
    user_id,                    # type: str
    authorizer=None,            # type: Optional[str]
    pos_to_close_user=None,     # type: Optional[str]
    table_service=None,         # type: Optional[Any]
):
    # type: (...) -> bool

    from table_actions import considers_tef_on_bordereau

    if pos_to_close_user is None:
        pos_to_close_user = pos_id

    _, _, session = get_operator_info(
        model=model,
        operator_id=user_id,
    )
    pod_function = sysactions.get_posfunction(model=model)
    current_order = sysactions.get_current_order(model=model)

    pod_type = sysactions.get_cfg(pos_id).key_value("pod") or "FL"
    is_quick_service = pod_type in ["FC", "FL"]

    if current_order is not None:
        table_id = current_order.find("CustomOrderProperties/OrderProperty/[@key='TABLE_ID']")
        is_quick_service = table_id is None

        if not close_order_from_another_pos(
            model=model,
            pos_id=pos_id,
            user_id=user_id,
            pos_to_close_user=pos_to_close_user,
            is_quick_service=is_quick_service,
            table_service=table_service,
        ):
            return False

    period = sysactions.get_business_period(model=model)

    from table_actions import bordereau_enabled

    report_name = "close_operator_report"

    if pod_function != "OT":
        response = do_transfer(
            pos_id=pos_id,
            transfer_type=6,
            operator_id=user_id,
            authorizator=authorizer,
            pos_to_close_user=pos_to_close_user,
        )
        done_transfer = response['done_transfer']
        need_ask_justification = response['need_ask_justification']
        set_authorizer(pos_id, authorizer, is_quick_service)
        if not done_transfer:
            return False

        if need_ask_justification:
            sysactions.print_report(
                pos_id,
                None,
                True,
                report_name,
                pos_to_close_user,
                period,
                user_id,
                session,
                authorizer,
                bordereau_enabled,
                considers_tef_on_bordereau,
            )

            account_repository = AccountRepository(mb_context=mb_context)
            get_justification(
                account_repository=account_repository,
                pos_id=pos_id,
                session=session,
            )

    logger.debug("Logoff POS %d - Deslogando operador %s" % (int(pos_id), user_id))

    if close_user(
        pos_id=pos_id,
        user_id=user_id,
        pos_to_close_user=pos_to_close_user,
    ):
        set_authorizer(
            pos_id=pos_id,
            authorizer=authorizer,
            is_quick_service=is_quick_service
        )
        logger.debug("Logoff POS %d - Operador %s deslogado com sucesso" % (int(pos_to_close_user), user_id))
    else:
        logger.error("Logoff POS %d - Falha ao deslogar operador %s" % (int(pos_to_close_user), user_id))
        sysactions.show_messagebox(pos_id, "$OPERATION_FAILED", "$ERROR")
        return False

    sysactions.print_report(
        pos_id,
        None,
        True,
        report_name,
        pos_to_close_user,
        period,
        user_id,
        session,
        authorizer,
        bordereau_enabled,
        considers_tef_on_bordereau,
    )

    return True


def set_authorizer(pos_id, authorizer, is_quick_service):
    if authorizer is not None and is_quick_service is False:
        sysactions.set_custom(pos_id, 'Last Manager ID', authorizer)
        sysactions.set_custom(pos_id, 'Authorization Level', get_user_level(authorizer))



import sys
import os
import iso8601
import numpy as np

from typing import List
from os import listdir
from os.path import isfile, join


class LogLine(object):
    def __init__(self, thread_number, log_date, message, line_content):
        self.thread_number = thread_number
        self.log_date = log_date
        self.message = message
        self.line_content = line_content


def read_log_files(log_files):
    file_data = ""
    for file_name in log_files:
        with open(file_name, "rb") as f:
            file_data += f.read()

    return file_data.decode("utf-8")


def get_thread_number(line):
    return line.split("-")[7].strip()


def get_message(line_content):
    index = line_content.index("MessageProcessorThread")
    index2 = line_content.index("$", index)
    return line_content[index2 + 1:]


def get_log_date(line_content):
    return iso8601.parse_date(line_content[0:23])


def get_log_lines(log_data):
    lines = []
    line_content = ""
    first_line = None
    for line in log_data.splitlines():
        if "MessageProcessorThread" not in line:
            continue

        line_start = line[0:23]
        try:
            iso8601.parse_date(line_start)
            if line_content != "":
                lines.append(
                    LogLine(
                        get_thread_number(first_line),
                        get_log_date(first_line),
                        get_message(line_content),
                        line_content)
                )
                line_content = ""

            first_line = line
        except Exception as _:  # noqa
            continue

        line_content += line + "\r\n"

    if line_content != "":
        try:
            lines.append(
                LogLine(
                    get_thread_number(first_line),
                    get_log_date(first_line),
                    get_message(line_content),
                    line_content)
            )
        except Exception as _:  # noqa
            pass

    return lines


def separate_lines_by_thread(log_lines):
    # type: (List[LogLine]) -> List[List[LogLine]]
    line_dict = {}
    for line in log_lines:
        if line.thread_number not in line_dict:
            line_dict[line.thread_number] = []
        line_dict[line.thread_number].append(line)

    return line_dict.values()


def get_connection_id(message):
    index = message.index("c_void_p(")
    index2 = message.index(")", index)
    return message[index + 9:index2]


def main():
    log_directory = os.path.abspath(sys.argv[1])
    log_files = [os.path.join(log_directory, f) for f in listdir(log_directory) if isfile(join(log_directory, f)) and "TimeActions.log" in f]
    log_lines = get_log_lines(read_log_files(log_files))
    separated_lines = separate_lines_by_thread(log_lines)

    message_times = {}
    message_max_times = {}
    start_time = None
    message = None
    start_line = None

    for thread_logs in separated_lines:
        for line in thread_logs:
            line.message = line.message.strip()

            if "MB_ReplyMessage" not in line.message and start_time is None:
                start_time = line.log_date
                start_line = line
                try:
                    message = line.message.split("$")[0]
                except Exception as _:  # noqa
                    start_time = None

            elif "MB_ReplyMessage" in line.message:
                end_time = line.log_date

                if end_time is not None and start_time is not None:
                    elapsed_time = (end_time - start_time).total_seconds()
                    if elapsed_time > 1000:
                        print(start_line)
                        start_time = None
                        continue

                    if elapsed_time > 0:
                        if message not in message_times:
                            message_times[message] = []
                        message_times[message].append(elapsed_time)

                        if message not in message_max_times or elapsed_time > message_max_times[message]:
                            message_max_times[message] = elapsed_time
                start_time = None

    messages = message_times.keys()
    messages.sort()
    for message in messages:
        times = message_times[message]
        a = np.array(times)
        p = np.percentile(a, 95)  # return 95th percentile

        ret_message = "{:40} - [{:>5}]: {:>6.3f} / {:>6.3f} / {:>6.3f}"
        print(ret_message.format(message, len(times), np.average(times), p, message_max_times[message]))
    print("")


main()



import json

from logging import Logger
from typing import (
    Any,
    Optional,
    Union,
)

from helper import remove_accents
from posot import OrderTaker

from model.config import Config
from model.repository import PosRepositoryModel
from model.services import (
    OrderServiceModel,
)
from model.error_type import PosValidationErrorsType
from model.exception import ValidationError
from model.order_detail import OrderDetail
from interactor.route.order.apply_benefit.dto import (
    ApplyBenefitRequest,
)
from validation_util import (
    check_if_pos_id_is_valid,
    validate_api_key,
    check_if_order_is_in_progress_or_totaled,
)


class ApplyBenefitInteractor(object):
    def __init__(
        self,               # type: ApplyBenefitInteractor
        logger,             # type: Logger
        config,             # type: Config
        pos_repository,     # type: PosRepositoryModel
        order_service,      # type: OrderServiceModel
    ):
        # type: (...) -> None

        self.logger = logger
        self.config = config
        self.pos_repository = pos_repository
        self.order_service = order_service
        self.model = None
        self.pos_ot = None  # type: Optional[OrderTaker]

    def execute(
        self,       # type: ApplyBenefitInteractor
        request,    # type: ApplyBenefitRequest
    ):
        # type: (...) -> Optional[Union[OrderDetail, str]]

        pos_id = validate_api_key(
            request=request,
            secure_api_key_list=self.config.api_key_mapping,
        )
        check_if_pos_id_is_valid(
            repository=self.pos_repository,
            pos_id=pos_id,
        )
        self.model = self.pos_repository.get_model(pos_id=pos_id)
        self.pos_ot = self.pos_repository.get_pos_ot(model=self.model)

        if self.pos_repository.has_current_order(model=self.model):
            check_if_order_is_in_progress_or_totaled(
                repository=self.pos_repository,
                model=self.model,
            )

        user_auth_level = int(self.model.find("Operator[@state='LOGGEDIN']").get("level") or -1)
        applied_benefit_dto = self.pos_repository.apply_benefit(
            pos_id=pos_id,
            benefit=request.benefit,
            user_auth_level=user_auth_level,
        )

        if applied_benefit_dto.benefit_applied:
            self.logger.info("Success benefit applied")
            order = self.pos_repository.get_current_order(model=self.model)
            order_id = order.get('orderId')
            order = self.pos_repository.get_order_details(
                pos_id=pos_id,
                order_id=order_id,
                pos_ot=self.pos_ot,
            )
            if order:
                return self.order_service.convert_mw_order_to_order_detail(order=order)

            return applied_benefit_dto.response_detail

        else:
            self._raise_custom_error(benefit_data=applied_benefit_dto.response_detail)

    def _raise_custom_error(
        self,
        benefit_data,    # type: Any[Union[str, json]]
    ):
        # type: (...) -> None

        if '\0' in benefit_data:
            error_code, message = benefit_data.split('\0')
        else:
            error_code = benefit_data
            message = ""

        self.logger.info("Error applying benefit. BenefitResponse: {}|{}".format(error_code, message))

        raise ValidationError(
            PosValidationErrorsType.CannotApplyBenefit,
            message,
        )




#
# The Python Imaging Library.
# $Id$
#
# sequence support classes
#
# history:
# 1997-02-20 fl     Created
#
# Copyright (c) 1997 by Secret Labs AB.
# Copyright (c) 1997 by Fredrik Lundh.
#
# See the README file for information on usage and redistribution.
#

##


class Iterator(object):
    """
    This class implements an iterator object that can be used to loop
    over an image sequence.

    You can use the ``[]`` operator to access elements by index. This operator
    will raise an :py:exc:`IndexError` if you try to access a nonexistent
    frame.

    :param im: An image object.
    """

    def __init__(self, im):
        if not hasattr(im, "seek"):
            raise AttributeError("im must have seek method")
        self.im = im
        self.position = getattr(self.im, "_min_frame", 0)

    def __getitem__(self, ix):
        try:
            self.im.seek(ix)
            return self.im
        except EOFError:
            raise IndexError  # end of sequence

    def __iter__(self):
        return self

    def __next__(self):
        try:
            self.im.seek(self.position)
            self.position += 1
            return self.im
        except EOFError:
            raise StopIteration

    def next(self):
        return self.__next__()


def all_frames(im, func=None):
    """
    Applies a given function to all frames in an image or a list of images.
    The frames are returned as a list of separate images.

    :param im: An image, or a list of images.
    :param func: The function to apply to all of the image frames.
    :returns: A list of images.
    """
    if not isinstance(im, list):
        im = [im]

    ims = []
    for imSequence in im:
        current = imSequence.tell()

        ims += [im_frame.copy() for im_frame in Iterator(imSequence)]

        imSequence.seek(current)
    return [func(im) for im in ims] if func else ims



from typing import Optional


class ElectronicMessage(object):

    def __init__(self, pos_id, event_type, timestamp, message_id=None, title=None, message_data=None, code_type=None,
                 input_type=None, choices=None, buttons=None):
        # type: (int, str, str, Optional[int], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str], Optional[str]) -> None

        self.pos_id = pos_id
        self.event_type = event_type
        self.timestamp = timestamp
        self.message_id = message_id
        self.title = title
        self.message_data = message_data
        self.code_type = code_type
        self.input_type = input_type
        self.choices = choices
        self.buttons = buttons



# -*- coding: utf-8 -*-
import json
import datetime
import logging
import os
import time

import cfgtools
from functools import wraps
from typing import (
    List,
    Optional,
    Union
)
from xml.etree import cElementTree as eTree

import persistence
import sysactions
from typing import (
    Any,
    Optional
)

import reportactions

from actions.coupon.display_coupons_from_report import display_coupons_from_report
from actions.login.utils import get_user_info
from actions.models import TransferType
from old_helper import config_logger, convert_from_utf_to_localtime
from msgbus import TK_SYS_ACK, TK_PRN_PRINT, FM_PARAM, MBException
from sysactions import get_model, show_keyboard, is_valid_date, show_info_message, action, \
    get_user_information, generate_report, show_listbox, show_messagebox, print_text, sys_log_info

from actions.util import get_opened_pos_user
from actions.util.get_menu_user_logged import get_menu_manager_authenticated_user
from systools import sys_log_exception

from utilfunctions import get_device_service
from xml.etree import cElementTree as eTree

mbcontext = None
poslist = []
store_id = None
config_logger(os.environ["LOADERCFG"], 'ManagerReports')
report_logger = logging.getLogger("ManagerReports")
config_logger(os.environ["LOADERCFG"], 'ManagerReportsData')
report_data_logger = logging.getLogger("ManagerReportsData")

invalid_date_i18n = "$INVALID_DATE"
type_report_i18n = "$TYPE_REPORT_POS"
type_operator_id_i18n = "$TYPE_OPERATOR_ID"
invalid_user_id_i18n = "$INVALID_USER_ID"
invalid_date_future_i18n = "$INVALID_DATE_FUTURE"
invalid_date_max_period_i18n = "$INVALID_DATE_MAX_PERIOD"
invalid_date_range_i18n = "$INVALID_DATE_RANGE"


def report_action(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        try:
            report_logger.info('Requested report: {}'.format(f.func_name))
        except Exception as _: # noqa
            pass

        report = f(*args, **kwargs)
        if report:
            report_data_logger.info(
                "\n".join((
                    '\n',
                    '********************************************************************',
                    'Requested report: {}'.format(f.func_name),
                    'Args: {} | Kwargs: {}'.format(str(args), str(kwargs)),
                    str(report),
                    '********************************************************************'
                ))
            )
        return report

    return action(decorated_function)


def print_report(pos_id, model, preview, report_name, *report_params):
    report = generate_report(report_name, *report_params)
    if report:
        current_printer = get_device_service(
            pos_id=pos_id,
            device_type="printer",
        )
        print_text(
            posid=pos_id,
            model=model,
            text=report,
            preview=preview,
            force_printer=current_printer,
        )
        return report
    else:
        sys_log_info("Could not generate report: %s" % report_name)
    return None


@report_action
def speedOfServiceReport(pos_id, store_wide="false", *args):
    model = get_model(pos_id)

    period = show_keyboard(pos_id, "$TYPE_DATE", title="", mask="DATE", numpad=True)
    if period is None:
        return

    if not is_valid_date(period):
        show_messagebox(pos_id, invalid_date_i18n)
        return

    if is_date_old(period):
        show_messagebox(pos_id, "$PERIOD_GREATER_THAN_THIRTY_DAYS")
        return

    pos = show_keyboard(pos_id, type_report_i18n, title="", defvalue="0", mask="INTEGER", numpad=True)
    if pos in (None, "", "."):
        return

    return print_report(pos_id, model, True, "speedOfService", pos_id, period, pos, store_id)


@report_action
def hourlySalesReport(
    pos_id,                 # type: str
    ask_pos="ask",          # type: Optional[str]
    ask_operator="ask",     # type: Optional[str]
    operator_id=None,       # type: Optional[str]
    releases=False,         # type: Optional[bool]
):
    # type: (...) -> Any

    from msgbusboundary import MsgBusTableService as tableService

    model = get_model(pos_id)
    period = show_keyboard(pos_id, "$TYPE_DATE", title="", mask="DATE", numpad=True)
    if period is None:
        return
    if not is_valid_date(period):
        show_messagebox(pos_id, invalid_date_i18n)
        return

    if is_valid_period(period, pos_id) is False:
        return None

    if is_date_old(period):
        show_messagebox(pos_id, "$PERIOD_GREATER_THAN_THIRTY_DAYS")
        return

    pos_list = []
    if ask_pos == "ask":
        pos_list = show_keyboard(pos_id, type_report_i18n, title="", defvalue="0", mask="INTEGER", numpad=True)
        if pos_list in (None, "", "."):
            return

    pos_list = tableService().get_exclusive_pos_list() if int(pos_list) == 0 else [int(pos_list)]
    if ask_operator == "ask":
        operator_id = show_keyboard(pos_id, type_operator_id_i18n, title="", defvalue="0", mask="INTEGER", numpad=True)
        if operator_id in (None, "", "."):
            return

        if operator_id == "0":
            operator_id = ""
        else:
            if not get_user_information(operator_id):
                show_info_message(pos_id, invalid_user_id_i18n, msgtype="warning")
                return

            operator_id = int(operator_id)

    return print_report(
        pos_id,
        model,
        True,
        "hourlySales",
        pos_id,
        json.dumps(pos_list),
        str(operator_id),
        period,
        releases
    )


def _get_report_parameters(
    pos_id,                             # type: Union[str, int]
    model,                              # type: eTree.XML
    ask_operator,                       # type: str
    date_type,                          # type: str
    ask_pos,                            # type: Union[str, int]
    operator_id=None,                   # type: Optional[Union[str, int]]
    ask_date=True,                      # type: Optional[bool]
    ignore_operator_not_found=False,    # type: Optional[bool]
):
    # type: (...) -> List

    if date_type == "SessionId":
        operator = model.find("Operator")
        if operator is None or operator.get("state") != "LOGGEDIN":
            show_messagebox(pos_id, "$USER_NOT_OPENED")
            raise sysactions.StopAction()

        session_id = model.find("Operator").get("sessionId")
        initial_date = None
        end_date = None
        operator_id = None
        report_pos = 0

    else:
        session_id = None
        initial_date = get_report_date(pos_id, "$TYPE_INITIAL_DATE") if ask_date else None
        if ask_date and initial_date is None:
            raise sysactions.StopAction()

        end_date = get_report_date(pos_id, "$TYPE_FINAL_DATE") if ask_date else None
        if ask_date and end_date is None:
            raise sysactions.StopAction()

        if initial_date and end_date and is_valid_date_range(initial_date, end_date, pos_id) is False:
            raise sysactions.StopAction()

        if ask_pos == "ask" or type(ask_pos) == int:
            if type(ask_pos) == int:
                report_pos = str(ask_pos)
            else:
                report_pos = show_keyboard(pos_id, type_report_i18n, title="", defvalue="0", mask="INTEGER",
                                           numpad=True)
                if report_pos in (None, "", "."):
                    raise sysactions.StopAction()
            report_pos = int(0 if report_pos in (None, 'None') else report_pos)
            if report_pos not in poslist and report_pos not in (0,):
                show_messagebox(pos_id, "$POS_NOT_FOUND")
                raise sysactions.StopAction()
            report_pos = str(report_pos)
        elif ask_pos == "current":
            report_pos = str(pos_id)
        else:
            report_pos = 0

        if ask_operator == "ask":
            operator_id = show_keyboard(
                posid=pos_id,
                message=type_operator_id_i18n,
                title="",
                defvalue="0",
                mask="INTEGER",
                numpad=True,
            )
            if operator_id in (None, "", "."):
                raise sysactions.StopAction()
            elif (not ignore_operator_not_found
                    and not get_user_information(user_number=operator_id)):
                show_messagebox(
                    posid=pos_id,
                    message=invalid_user_id_i18n,
                )
                raise sysactions.StopAction()

        elif ask_operator == "current":
            try:
                current_operator_id = model.find('Operator').get('id')
                operator_id = current_operator_id
            except Exception as _:
                show_messagebox(
                    posid=pos_id,
                    message="$USER_NOT_OPENED",
                )
                raise sysactions.StopAction()

    return [date_type, pos_id, report_pos, initial_date, end_date, operator_id, session_id]


@report_action
def productMixReportByPeriod(
    pos_id,                         # type: str
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="ask",                  # type: Optional[str]
    ask_operator="ask",             # type: Optional[str]
    operator_id=False,              # type: Optional[str]
):
    # type: (...) -> bool

    model = get_model(pos_id)

    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
    )
    parameters.extend([store_id, get_pmix_report_model(pos_id)])

    return print_report(pos_id, model, True, "new_pmix_report", *parameters)


@report_action
def tipReport(
    pos_id,                 # type: str
    ask_operator="ask",     # type: Optional[str]
    operator_id=None,       # type: Optional[str]
):
    # type: (...) -> bool

    model = get_model(pos_id)

    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type="BusinessPeriod",
        ask_pos="notAsk",
        operator_id=operator_id,
    )
    parameters.extend([store_id])

    return print_report(pos_id, model, True, "tip_report", *parameters)


@report_action
def cashReport(
    pos_id,                         # type: str
    ask_operator="ask",             # type: Optional[str]
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="ask",                  # type: Optional[str]
    is_table_service=False,         # type: Optional[bool]
    operator_id=None,               # type: Optional[str]
    only_body=False,                # type: Optional[bool]
    only_payments=False,            # type: Optional[bool]
    is_integrated_delivery=False,   # type: Optional[bool]
):
    # type: (...) -> bool

    model = get_model(pos_id)

    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
    )
    parameters.extend([store_id, None, None, None, bool(is_table_service), only_body, only_payments, False,
                       is_integrated_delivery])
    return print_report(pos_id, model, True, "new_cash_report", *parameters)


@report_action
def sales_summary_report(
    pos_id,                         # type: str
    ask_operator="ask",             # type: Optional[str]
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="ask",                  # type: Optional[str]
    operator_id=None,               # type: Optional[str]
):
    # type: (...) -> bool

    model = get_model(pos_id)

    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
    )
    report_parameters = [pos_id, parameters[3], parameters[4], store_id]
    return print_report(pos_id, model, True, "sales_summary_report", *report_parameters)


@report_action
def tef_cancellation_report(
    pos_id,                         # type: str
    ask_operator="ask",             # type: Optional[str]
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="ask",                  # type: Optional[str]
    operator_id=None,               # type: Optional[str]
):
    # type: (...) -> None

    model = get_model(pos_id)
    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
    )
    report_parameters = [
        parameters[3],
        parameters[4]
    ]
    reportactions.print_report(
        pos_id,
        "tefcancellationreport",
        True,
        None,
        pos_id,
        *report_parameters
    )


@report_action
def benefit_report(
    pos_id,                         # type: str
    ask_operator="ask",             # type: Optional[str]
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="ask",                  # type: Optional[str]
    operator_id=None,               # type: Optional[str]
):
    # type: (...) -> None

    from msgbusboundary import MsgBusTableService as tableService

    model = get_model(pos_id)
    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
        ignore_operator_not_found=True,
    )
    initial_date = parameters[3]
    end_date = parameters[4]
    operator_id = parameters[5] if parameters[5] != '0' else None
    benefit_selected = json.dumps(display_coupons_from_report(pos_id, initial_date, end_date, operator_id))
    pos_list = json.dumps(tableService().get_exclusive_pos_list())
    parameters.extend([benefit_selected, store_id, None, None, None, pos_list])
    reportactions.print_report(
        pos_id,
        "benefitsreport",
        True,
        None,
        pos_id,
        *parameters
    )

@report_action
def voidedOrdersReport(
    pos_id,                         # type: str
    ask_operator="ask",             # type: Optional[str]
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="ask",                  # type: Optional[str]
    operator_id=None,               # type: Optional[str]
):
    # type: (...) -> bool

    model = get_model(pos_id)

    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
    )
    parameters.extend([store_id])

    return print_report(pos_id, model, True, "voided_orders_report", *parameters)


def get_report_date(pos_id, message):
    # type (str, str) -> str
    date = show_keyboard(pos_id, message, title="", mask="DATE", numpad=True)

    if date is None:
        return None

    if not is_valid_date(date):
        show_messagebox(pos_id, invalid_date_i18n)
        raise sysactions.StopAction()

    if is_valid_period(date, pos_id) is False:
        raise sysactions.StopAction()

    return date


def is_valid_period(date, pos_id):
    # type (str) -> bool
    chosen_date = datetime.datetime.strptime(date, '%Y%m%d').date()
    days_from_today = datetime.date.today() - datetime.timedelta(days=90)

    if chosen_date < days_from_today:
        show_messagebox(pos_id, invalid_date_max_period_i18n)
        return False

    if chosen_date > datetime.date.today():
        show_messagebox(pos_id, invalid_date_future_i18n)
        return False


def is_valid_date_range(initial_date, end_date, pos_id):
    # type (str, str) -> bool
    initial_date = datetime.datetime.strptime(initial_date, '%Y%m%d').date()
    end_date = datetime.datetime.strptime(end_date, '%Y%m%d').date()

    if end_date < initial_date:
        show_messagebox(pos_id, invalid_date_range_i18n)
        return False


def get_pmix_report_model(pos_id):
    ret = show_messagebox(pos_id, "$CHOOSE_PMIX_PRODUCT_TYPE", title="", buttons="$ALL_PRODUCTS|$PRICED|$UNPRICED")

    if ret == 1:
        return "PRICED"
    elif ret == 2:
        return "UNPRICED"
    else:
        return "ALL_PRODUCTS"


@report_action
def checkoutReport(posid, *args):
    def get_option(session):
        close_time = datetime.datetime.strptime(session['closeTime'], "%Y-%m-%d %H:%M:%S").strftime("%H:%M:%S")
        return "%s - %s" % (close_time, session['operatorName'])

    def get_operator_sessions(operatorid, period, posid):
        sessions = []
        options = []
        session_cols = ('openTime', 'closeTime', 'sessionId', 'operatorName')

        conn = None
        try:
            conn = persistence.Driver().open(mbcontext, dbname=str(posid))
            conn.transaction_start()
            cursor = conn.pselect("drawerChangeReport", startPeriod=period, endPeriod=period, instance=posid)
            for row in cursor:
                if row.get_entry("closeTime") and (
                        int(row.get_entry("operatorId")) == int(operatorid) or int(operatorid) == 0):
                    sessions.append(dict([(col, row.get_entry(col)) for col in session_cols]))
                    options.append(get_option(sessions[-1]))

            conn.transaction_end()
        finally:
            if conn:
                conn.close()
        return sessions, options

    model = get_model(posid)
    period = show_keyboard(posid, "$ENTER_DATE_OR_PRESS_OK", title="", mask="DATE", numpad=True)
    if period is None:
        return
    if not is_valid_date(period):
        show_info_message(posid, invalid_date_i18n, msgtype="error")
        return

    operatorid = show_keyboard(posid, type_operator_id_i18n, title="", defvalue="0", mask="INTEGER", numpad=True)
    if operatorid in (None, "", "."):
        return
    if not get_user_information(operatorid):
        show_info_message(posid, invalid_user_id_i18n, msgtype="warning")
        return

    sessions, options = get_operator_sessions(operatorid, period, posid)
    if sessions:
        index = show_listbox(posid, options)
        if index is None:
            return  # The user cancelled - or timed-out

        session_id = sessions[index]['sessionId']
        menu_manager_user = get_menu_manager_authenticated_user(int(posid))
        return print_report(posid, model, True, "cash", posid, period, operatorid, "False", "0", "Check Out Report",
                            session_id, menu_manager_user)
    else:
        show_messagebox(posid, message="$THERE_NO_SHIFT_OPERATOR")


def is_date_old(period, *args):
    consult_days = abs((datetime.datetime.strptime(period, "%Y%m%d") - datetime.datetime.now()).days)

    return consult_days > 30


def is_current_date(period, *args):
    return datetime.datetime.strptime(period, "%Y%m%d").date() == datetime.date.today()


@report_action
def transfers_report(pos_id, *args):
    model = get_model(pos_id)
    options = [sysactions.translate_message(model, "CASH_IN"), sysactions.translate_message(model, "CASH_OUT")]
    transfer_type = show_listbox(pos_id, options, message="$SELECT_OPTION")
    if transfer_type is None:
        return
    transfer_type = int(TransferType.TRANSFER_CASH_IN) if transfer_type == 0 else int(TransferType.TRANSFER_CASH_OUT)

    period = show_keyboard(pos_id, "$ENTER_DATE_OR_PRESS_OK", title="", mask="DATE",
                           numpad=True)
    if period is None:
        return
    if not is_valid_date(period):
        show_messagebox(pos_id, message=invalid_date_i18n, icon="error")
        return

    if is_valid_period(period, pos_id) is False:
        return

    report = generate_report("transfer_report_by_period", pos_id, "", transfer_type, period)
    reports = json.loads(report)
    data = []
    index = 0
    for report in reports:
        report_id = index
        index += 1
        for key, value in report.items():
            data.append((report_id, key, value))

    if not data:
        show_messagebox(pos_id, message="$HAS_NO_ITEMS_TO_SHOW", icon="error")
        return

    transfer_title = '$SELECT_OPTION'

    if transfer_type == 3:
        transfer_title = '$SELECT_SUPPLY'

    elif transfer_type == 4:
        transfer_title = '$SELECT_BLEED'

    selected_transfer = sysactions.show_text_preview(
        posid=pos_id,
        texts=data,
        title=transfer_title,
        buttons='$CANCEL|$PRINT',
        defvalue='',
        onthefly=False,
        timeout=120000,
    )

    if selected_transfer is None:
        return

    if int(selected_transfer[0]) == 0:
        try:
            message = None
            index = 0
            for msg in data:
                try:
                    if index == int(selected_transfer[1]):
                        message = str(msg[2])
                    index += 1
                except:
                    continue

            if message is not None:
                printer = None
                try:
                    printer = get_device_service(
                        pos_id=pos_id,
                        device_type="printer",
                    )
                    msg = mbcontext.MB_EasySendMessage(printer, TK_PRN_PRINT, FM_PARAM, message)
                except MBException as _:
                    time.sleep(0.5)
                    msg = mbcontext.MB_EasySendMessage(printer, TK_PRN_PRINT, FM_PARAM, message)
                finally:
                    if msg.token == TK_SYS_ACK:
                        show_info_message(pos_id, "$OPERATION_SUCCEEDED", msgtype="success")
                    else:
                        show_info_message(pos_id, "$OPERATION_FAILED", msgtype="error")
        except Exception as _:
            show_info_message(pos_id, "$ERROR_PRINT_REPORT", msgtype="error")


@report_action
def opened_tables_report(pos_id):
    model = get_model(pos_id)
    return print_report(pos_id, model, True, "openedtablesreport", pos_id)


@report_action
def doPrintLogoffReport(pos_id):
    from table_actions import considers_tef_on_bordereau

    model = get_model(pos_id)
    sessions = []

    period = show_keyboard(pos_id, "$TYPE_DATE", title="", mask="DATE", numpad=True)

    if period in (None, ''):
        return

    conn = None
    try:
        date_input_format = "%Y-%m-%d %H:%M:%S"
        date_output_format = "%d/%m/%Y %H:%M:%S"

        conn = persistence.Driver().open(mbcontext, dbname=str(pos_id))
        conn.transaction_start()

        sql = """SELECT * FROM posctrl.UserSession
                  WHERE BusinessPeriod = '%s'
                  ORDER BY CloseTime DESC
                  """ % period
        cursor = conn.select(sql)

        for row in cursor:
            session_id = row.get_entry("SessionId")
            pos = session_id.split(",")[0].split("=")[1]
            close_time = row.get_entry("CloseTime")
            if close_time is not None:
                open_formatted_time = datetime.datetime.strptime(row.get_entry("OpenTime"), date_input_format)
                open_time = convert_from_utf_to_localtime(open_formatted_time).strftime(date_output_format)

                close_formatted_time = datetime.datetime.strptime(close_time, date_input_format)
                close_time = convert_from_utf_to_localtime(close_formatted_time).strftime(date_output_format)

                operator_name = row.get_entry("OperatorName")
                description = operator_name + "  " + open_time
                msg = "Loja: %s" \
                      "\nPOS: %s" \
                      "\nNome do operador: %s" \
                      "\nHora de entrada: %s" \
                      "\nHora de saída: %s" % (store_id, pos, operator_name, open_time, close_time)
                sessions.append((session_id, description, msg))

    except Exception as e:
        sys_log_exception('Exception printing logoff report: {}'.format(e))
        show_info_message(pos_id, "$OPERATION_FAILED", msgtype="error")
        return
    finally:
        if conn:
            conn.close()

    if not sessions:
        show_messagebox(pos_id, "$THERE_ARE_NO_CLOSED_SESSIONS", title="$ERROR", icon="error", buttons="$OK")
        return

    selected = sysactions.show_text_preview(
        posid=pos_id,
        texts=sessions,
        title='$SELECT_AN_OPTION',
        buttons='$CANCEL|$PRINT',
        defvalue='',
        onthefly=False,
        timeout=120000
    )

    if selected is not None and selected[0] == '1':
        try:
            period = selected[4][7:15]
            userid = selected[2][5:]
            session_id = selected[1] + "," + selected[2] + "," + selected[3] + "," + selected[4]
            select_pos_id = selected[1][4:]

            from table_actions import bordereau_enabled
            report_name = "close_operator_report"
            return print_report(
                pos_id,
                model,
                True,
                report_name,
                select_pos_id,
                period,
                userid,
                session_id,
                "",
                bordereau_enabled,
                considers_tef_on_bordereau,
            )
        except Exception as e:
            show_info_message(pos_id, "$ERROR_TRYING_TO_REPRINT_REPORT", msgtype="error")


@report_action
def get_voided_lines_report(pos_id):
    start_date = get_report_date(pos_id, "$TYPE_INITIAL_DATE")
    if start_date is None:
        raise sysactions.StopAction()
    end_date = get_report_date(pos_id, "$TYPE_FINAL_DATE")
    if end_date is None:
        raise sysactions.StopAction()

    if is_valid_date_range(start_date, end_date, pos_id) is False:
        return

    report_pos = sysactions.show_keyboard(pos_id, "$TYPE_REPORT_POS", "", defvalue="0", mask="INTEGER", numpad=True)
    if report_pos is None:
        return

    operator_id = sysactions.show_keyboard(pos_id, "$TYPE_OPERATOR_ID", "", defvalue="0", mask="INTEGER", numpad=True)
    if operator_id is None:
        return

    model = sysactions.get_model(pos_id)

    report_name = "generate_voided_lines_report"
    return print_report(pos_id, model, True, report_name, report_pos, start_date, end_date, False, operator_id)


def sale_challenge_report(
    pos_id,                 # type: Any[Union[str, int]]
    sale_challenge_id,      # type: int
    ask_operator="ask",     # type: Optional[str]
    is_manager=False,       # type: Optional[bool]
    operator_id=None,       # type: Optional[str]
):
    # type: (...) -> Optional[bool]

    model = get_model(pos_id)

    ask_operator = ask_operator if is_manager else "current"
    operator_id = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type="",
        ask_pos="",
        operator_id=operator_id,
        ask_date=False,
    )[5]
    return reportactions.print_report(
        pos_id,
        "sale_challenge_report",
        True,
        None,
        pos_id,
        operator_id,
        sale_challenge_id
    )


@report_action
def show_justification_discount_report(
    pos_id,                         # type: str
    ask_operator="ask",             # type: str
    operator_id=None,               # type: Optional[str]
    date_type="BusinessPeriod",     # type: Optional[str]
    ask_pos="current",              # type: Optional[str]
):
    # type: (...) -> None

    model = get_model(
        posid=pos_id,
    )

    parameters = _get_report_parameters(
        pos_id=pos_id,
        model=model,
        ask_operator=ask_operator,
        date_type=date_type,
        ask_pos=ask_pos,
        operator_id=operator_id,
    )

    initial_date = parameters[3]
    end_date = parameters[4]
    operator_id = parameters[5] if parameters[5] != '0' else None

    parameters.extend([initial_date, end_date, operator_id])
    reportactions.print_report(
        pos_id,
        "justificationdiscountreport",
        True,
        None,
        pos_id,
        *parameters
    )


def justification_report(
    pos_id,                                 # type: str
    justification,                          # type: str
    discount_description=None,              # type: Optional[str]
    justification_type="APPLY_DISCOUNT",    # type: Optional[str]
    authorized_by=None,                     # type: Optional[str]
    order_id=None,                          # type: Optional[str]
):
    # type: (...) -> None

    if justification in ['', None]:
        return

    if not _has_config_print_justification(
        justification_type=justification_type,
    ):
        return

    model = get_model(pos_id)
    operator = model.find("Operator")
    operator_id = operator.get("id") if operator is not None else None

    authorized_by = _get_user_authorization(
        operator_id=operator_id,
        authorized_by=authorized_by,
    )

    if order_id is None:
        order = sysactions.get_current_order(
            model=model,
        )
        order_id = order.get("orderId")

    data = {
        "orderId": order_id,
        "operator": operator_id,
        "AuthorizedBy": authorized_by,
        "justificationType": justification_type,
        "discountUserDescription": discount_description,
        "justification": justification,
    }

    data = json.dumps(
        obj=data,
    )

    reportactions.print_report(
        pos_id,
        "justificationreport",
        False,
        None,
        data,
    )


def _get_user_authorization(
    operator_id,           # type: str
    authorized_by=None,    # type: Optional[str]
):
    if authorized_by is not None:
        return authorized_by

    user_xml_str = get_user_information(
        user_number=operator_id,
    )

    if user_xml_str is None:
        return operator_id

    user_element = eTree.XML(
        text=user_xml_str
    ).find("user")

    return user_element.attrib["UserName"]


def _has_config_print_justification(
    justification_type="APPLY_DISCOUNT",    # type: str
):
    # type: (...) -> bool

    config = cfgtools.read(
        file_or_path=os.environ["LOADERCFG"],
    )

    if justification_type == "APPLY_DISCOUNT":
        return str(config.find_value(
            path="Customizations.PrintApplyDiscountJustifyCoupon"
        ) or 'false'.lower()) == "true"

    order_state_cancel_print = config.find_values(path="Customizations.PrintCancelJustifyCoupon")

    if order_state_cancel_print is None:
        return False

    return justification_type in order_state_cancel_print




from messagebus import TokenCreator, TokenPriority


class BusTokens(object):

    # RemoteOrder
    TK_REMOTE_ORDER_GET_RUPTURED_ITEMS = TokenCreator.create_token(TokenPriority.low, "37", "213")

    # FiscalWrapper
    TK_FISCALWRAPPER_PROCESS_REQUEST = TokenCreator.create_token(TokenPriority.high, "31", "1")
    TK_FISCALWRAPPER_SEFAZ_CONNECTIVITY = TokenCreator.create_token(TokenPriority.high, "31", "4")
    TK_FISCALWRAPPER_SAT_OPERATIONAL_STATUS_REQUEST = TokenCreator.create_token(TokenPriority.high, "31", "5")
    TK_FISCALWRAPPER_GET_NF_TYPE = TokenCreator.create_token(TokenPriority.low, "31", "7")
    TK_FISCALWRAPPER_SAT_STORE_STATUS = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="31",
        number="32",
    )

    # EletronicPayment
    TK_ELECTRONIC_PAYMENT_ADD_PAYMENT = TokenCreator.create_token(TokenPriority.low, "0E", "1")
    TK_ELECTRONIC_PAYMENT_CANCEL_PAYMENT = TokenCreator.create_token(TokenPriority.low, "0E", "2")
    TK_ELECTRONIC_PAYMENT_FINISH_PAYMENT = TokenCreator.create_token(TokenPriority.low, "0E", "3")
    TK_ELECTRONIC_PAYMENT_CONNECTIVITY_TEST = TokenCreator.create_token(TokenPriority.low, "0E", "4")
    TK_ELECTRONIC_PAYMENT_SEND_RESPONSE = TokenCreator.create_token(TokenPriority.low, "0E", "5")
    TK_ELECTRONIC_PAYMENT_GET_ELECTRONIC_TYPES = TokenCreator.create_token(TokenPriority.low, "0E", "8")

    # Benefits
    TK_NDISCOUNT_ADD_BENEFIT = TokenCreator.create_token(TokenPriority.low, "112", "1")

    # Coupons
    TK_COUPON_GET = TokenCreator.create_token(TokenPriority.low, "46", "2")
    TK_COUPON_GET_FREE_COUPON_DATA = TokenCreator.create_token(TokenPriority.low, "46", "B")
    TK_COUPON_APPLY = TokenCreator.create_token(TokenPriority.low, "113", "1")
    TK_COUPON_REMOVE = TokenCreator.create_token(TokenPriority.low, "112", "2")

    # PosUpdater
    TK_POS_UPDATER_GET_VERSIONS = TokenCreator.create_token(TokenPriority.low, "95", "6")

    # I18N
    TK_I18N_GETLANGS = TokenCreator.create_token(TokenPriority.low, "7", "4")
    TK_I18N_GETTABLEJSON = TokenCreator.create_token(TokenPriority.low, "7", "5")

    # Static Config
    TK_POS_EXECUTEACTION = TokenCreator.create_token(TokenPriority.high, "5", "3")

    # Device Manager
    TK_GET_USED_SERVICE = TokenCreator.create_token(TokenPriority.low, "D", "1")

    # NSale/ProductManager
    TK_SLCTRL_PMGR_GET_NAVIGATION = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="A",
        number="61",
    )

    TK_SLCTRL_PMGR_GET_MAIN_MENU = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="A",
        number="64",
    )

    # Loyalty
    TK_LOYALTY_GET_LOYALTY_ID = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="111",
        number="1",
    )
    TK_LOYALTY_GET_AND_LOCK_VOUCHER = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="111",
        number="2",
    )
    TK_LOYALTY_BURN_VOUCHER = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="111",
        number="3",
    )
    TK_LOYALTY_UNLOCK_VOUCHER = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="111",
        number="4",
    )
    TK_LOYALTY_IS_AVAILABLE = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="111",
        number="FF",
    )
    TK_LOYALTY_GET_ID_TYPE = TokenCreator.create_token(
        priority=TokenPriority.low,
        message_group="111",
        number="5",
    )



from model.log import LogEntry


class EventDispatch(object):
    def __init__(self, event_time, row_id, log_entry, encoded=None):
        # type: (str, int, LogEntry, bool) -> None
        self.event_time = event_time
        self.row_id = row_id
        self.log_entry = log_entry
        self.encoded = encoded
